#!/usr/bin/perl

# Program to get/put data from/to an IC-91AD
# Author: Bruce Hunsaker - kg7wi@arrl.net
#
# Copyright (C) 2007 Bruce Hunsaker. All Rights Reserved.
#
#   This is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This software is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
# TODO:
#   Get data from radio:
#   1) Done. Get memory data for band A and band B
#   2) Done. Get Programmed scan data for band A and band B
#   3) Done. Get Call channel data for band A and band B
#   4) Get VFO data for band A and band B (Not sure this is possible via data port)
#   5) Done. Output data to a .csv file
#   Update radio:
#   6) Done. Read .csv file
#   7) Done. Put memory data for band A and band B
#   8) Done. Put Programmed scan data for band A and band B
#   9) Done. Put Call channel data for band A and band B
#  10) Put VFO data for band A and band B (Not possible via data port?)
#      Try 1a 00 04-07
#  11) Done. Allow entries in .csv to clear memory channels. Use -1 as the frequency.
#
# Advanced functionality:
#   1) Done. Get/put my call sign data
#   2) Done. Get Received call sign data (20)
#      fe fe 02 80 1d 04 nn fd
#      fe fe 80 02 1d 04 nn xx <rpt2> <rpt1> <called> <caller> </> fd
#      nn = rx call record number (binary>
#      xx = status
#      <rpt2> 8 bytes
#      <rpt1> 8 bytes
#      <called> 8 bytes
#      <caller> 8 bytes
#      </> 4 bytes
#   3) Done. Get/put TX messages (5)
#   4) Done. Get/put bank-name data (A & B) (A-Z)
#   5) Done. Get/put bank-link data (A & B)
#   6) Done. Add command line option(s) to get/put only specified memories.
#      For example get/put only mycall, or get/put only bank A.
#      --get=all,a,b,mycall,txmsg,banknm[ab],bankln[ab]
#      --put=all,a,b,mycall,txmsg,banknm[ab],bankln[ab]
#   7) Add command line option to output lines for all memories even
#      if they are clear.
#   8) Done. Get/put Your Call Sign data (60 memories)
#      fefe 0280 1d 06 01 <call sign>{8} fd - 01 is memory number (binary 0x28 = 40)
#   9) Done. Get/put Repeater Call Sign data (60 memories)
#      fefe 0280 1d07 02 <callsign>{8} fd  (for memory 2)
#  10) Get/put VFO Call sign data (2 memories VHF and UHF)
#      1a 00 00 07 00 00 70 48 04 00 00 00 05 00 25 41
#            25 41 04 65 22 80 46 01 00 00 51 
#      Note VFO memory data is in a different format!
#  11) Done. Get/put DTMF memory (10 memories)
#      1a06 02 <number>{up to 16 ascii characters} fd - For memory #2
#      0-9,A-D,#,*,
#      Selected DTMF memory: 1a 07 nn fd
#  12) Done. Put '*' instead of memory channel number. When loading it uses the
#      next one in sequence (like DStarCom). If channel number starts with '*'
#      then the next sequential one will be used when loading.  The .csv file
#      contains '* <chno>' by default so the current channel number is
#      shown. However <chno> is optional when loading back to the radio.
#  13) Done. Get/Put TV channel data.
#      TV channel data contains:
#       a) frequency
#       b) Mode (WFM or AM)
#       c) Skip (Skip or OFF)
#       d) Channel ID 00-JJ
#
#      Data format:
#        fe fe 01 80 1a 00 03 02 fd    |             <- Readout request
#        fe fe 80 01 1a 00 03 02 00 00 |             <- Readout data
#        75 71 00 04 20 34 fd          | uq   4    
#
#        5 byte frequency (BCD LSB -> MSB) above example is:
#           0071.750000 Mhz
#      
#        1 byte indicating skip and mode
#          04 WFM OFF
#          05 WFM Skip
#          06 AM  OFF
#          07 AM  Skip
#          
#          This might imply:
#            00 FM   OFF
#            01 FM   Skip
#            02 N-FM OFF
#            03 N-FM Skip
#          Note: Radio does not accept FM, FM-N or DV as Modes for TV channels.
#                (Too bad)
#
#        2 bytes for ascii channel number (' 4' in above example)
#          Valid characters seem to be '0123456789ABCDEFGHIJ'
#
#       It would be interesting to see what other uses for this memory could be
#       since it becomes obsolete in Feb 2009.
#       Here are some Ideas (if they work):
#         a) Top 70 Shortwave Frequencies
#         b) A mix of AM and FM Broadcast stations.
#         c) Air frequencies (Since these are AM)
#
#         It appears that the tuning step is 25 Khz, this means
#         I can put in 1.150 or 1.175 but not 1.160. This will limit
#         which AM and shortwave frequencies I can use, but should work
#         OK for the Air bands and FM broadcast band.
#
#         Update: It looks like the 25 khz tuning step restriction
#         may be an RS-91 restriction. I had no trouble putting in 
#         1.160 AM using this program.
#
# More functionality
#   1) Get/Set settings (A & B)
#   2) Radio Operation (probably a different app.
#   3) Data communications app (texting, mail? BBS? forwarding? gateway?)
#
# NOTE: Prefixing text data with a single quite (') seems to fix the load problem
#       where the data might be taken as some other data type.
#       This works when loading into a spreadsheet as the quote is removed
#       when the file is saved.  Other editing methods will leave the quote
#       intact.

use strict;
use Getopt::Long;
use Text::ParseWords;

my $port = "/dev/ttyS0";
#my $port = "/tmp/interceptty";

my $debug = 0;

# operations
my $get = 0;
my $put = 0;

#
my $memgroup = "all";

$| = 1; #autoflush STDOUT
select((select(STDERR), $| = 1)[0]);  #autoflush STDERR

&GetOptions(
  "get" => \$get,
  "put" => \$put,
  "port=s" => \$port,
  "debug" => \$debug,
  "group=s" => \$memgroup
);
my $csvfile = shift;


#
# Must supply a filename, and either get or put (but not both)
# on the cmomand line.
if (! $csvfile || ($get && $put) || (!$get && !$put) ) {
  die "Usage: $0 [--get || --put] [--port=<portname>]\n" .
      "\t\t[--group=<memtype>[,<memtype>[...]] [--debug] <csvfile>\n\n" .
      "      --get   Reads data from the radio and outputs to <csvfile>.\n" .
      "      --put   Reads data from <csvfile> and writes it to the radio.\n" .
      "      --port  Uses the port <portname> as the serial port the radio\n" .
      "              is connected to.  Defaults to '/dev/ttyS0'.\n" .
      "      --group Specifies which memory groups to get/put.\n" .
      "              Values for <memtype> are:\n" .
      "                - all    all supported memory types.\n" .
      "                - bandA  band A memory channels\n" .
      "                - bandB  band B memory channels\n" .
      "                - mycall my call memories.\n" .
      "                - urcall your call memories.\n" .
      "                - rpcall repeater call memories.\n" .
      "                - banknm Bank Names\n" .
      "                - bankln Bank Link configuration.\n" .
      "                - txmsg  TX message memories.\n" .
      "                - dtmf   DTMF memories.\n" .
      "                - rxlog  Rx call records (get only).\n" .
      "                - tv     TV channel memories.\n" .
      "              If this option is not present, the default is 'all'.\n";
}

# memory groups to process.
my %mgopt;
$mgopt{bandA} = 1;
$mgopt{bandB} = 1;
$mgopt{mycall} = 1;
$mgopt{urcall} = 1;
$mgopt{rpcall} = 1;
$mgopt{banknm} = 1;
$mgopt{bankln} = 1;
$mgopt{txmsg} = 1;
$mgopt{rxlog} = 1;
$mgopt{dtmf} = 1;
$mgopt{tv} = 1;


# Commands:
my %cmd;

$cmd{interrogate} = "fefefefefefefefefefefefefefefefefefe018019fd";
$cmd{getmemApre} = pack("H*","fefe01801a0001"); #append mem channel
$cmd{getmemBpre} = pack("H*","fefe02801a0001"); #append mem channel

my @mod_mode = ("FM","FM-N","wFM","AM","DV");
my @tstep_code = (
    "5",
    "6.25",
    "7.5",
    "9",
    "10",
    "12.5",
    "15",
    "20",
    "25",
    "30",
    "50",
    "100",
    "125",
    "200"
  );

my @dup_code = ("OFF","Dup-","Dup+");

my %tone_code;
$tone_code{0x00} = "OFF";
$tone_code{0x04} = "TONE";
$tone_code{0x08} = "TSQLPB";
$tone_code{0x0c} = "TSQL";
$tone_code{0x10} = "DTCSPB";
$tone_code{0x14} = "DTCS";
$tone_code{0x18} = "TSQLR"; # Inversion tone squelch (Non-USA)
$tone_code{0x1c} = "TRAIN1"; # Mute if 'train freq' detected (empty-line squelch)
$tone_code{0x20} = "TRAIN2"; # MSK 1500Hz tone and single-switch control signal detection.
                             # An empty line is indicated by a fixed frequency signal.
                             # Alternate translation:
                             # Mute if  alternating single tone 
                             # and MSK 1500Hz control signal is detected.
$tone_code{0x24} = "MSK";    # Mute MSK 1200 bps Data (BIIS / MDC1200 ?)

my @skip_code = ("OFF","skip","Pskip");
my %pol_code;
$pol_code{0x00} = "NN";
$pol_code{0x04} = "NR";
$pol_code{0x08} = "RN";
$pol_code{0x0c} = "RR";

my %dsql_type;
$dsql_type{0x00} = "OFF";
$dsql_type{0x01} = "DBEEP";
$dsql_type{0x02} = "DSQL";
$dsql_type{0x03} = "CBEEP";
$dsql_type{0x04} = "CSQL";

my $last_band; #used by status display

#system("stty 38400 -raw -echo <$port");
system("stty 38400 -raw -echo min 1 -icanon -isig -ignpar -inpck -istrip -icrnl -ixon -ixoff -opost -iexten <$port");

open(RADIO,"+<$port") || die "Can't open port $port\n";

# main
&set_mgopt;
if ($get) {
  &get_data;
}
elsif ($put) {
  &put_data;
}

close(RADIO);
exit 0;

sub set_mgopt {
  if (length($memgroup) > 0) {
    my @opt = split(',',$memgroup);
    # User specified specific memrory banks to get.
    for my $key (keys %mgopt) {
      $mgopt{$key} = 0; #default to off.
      if ( grep(/^$key$/i,@opt) || $memgroup =~ /^all/i ) {
        $mgopt{$key} = 1;
      }
    }
  }
  if ($debug) {
    print STDERR "Getting memories: ";
    for my $key (sort keys %mgopt) {
      print STDERR " $key" if ( $mgopt{$key} );
    }
    print STDERR "\n";
  }
}

#
# Get radio ID information. We expect to see the ID for an IC-91AD
#
sub get_ID {
  # make sure we are talking, and talking to an IC-91A/D
  &send($cmd{interrogate});
  my $msg = &recv("fefe800119");
  #expecting: fefe80011928880100050f02017780fd
  my $model = substr($msg,10,6);
  if ($model eq "288801") {
    print STDERR "Connected\n";
  }
  else {
    die "Invalid radio model: $model\n";
  }
}

#
# Get a 'mycall' entry from the radio.
#
sub get_mycall {
  my $mem_no = shift;
  $mem_no = sprintf("%02d",$mem_no);
  my $cmd = "fefe02801d08${mem_no}fd";
  #my $cmd = "fefe02801d08fd";
  &send($cmd);
  my $msg = &recv("fefe80021d08${mem_no}");
  my $call = pack("H*",substr($msg,14,16));
  my $slash = pack("H*",substr($msg,30,8));
  my $call = "$call/$slash";
  return $call;
}

#
# Get the number of the selected 'mycall' memory.
#
sub get_sel_mycall {
  my $cmd = "fefe02801d02fd";
  &send($cmd);
  my $msg = &recv("fefe80021d02");
  my $sel_call = substr($msg,12,2);
  $sel_call += 0;
  return $sel_call;
}

#
# write a 'mycall' memory to the radio.
#
sub put_mycall {
  my $mem_no = shift;
  my $sel = shift;
  my $full_call = shift;
  $mem_no = sprintf("%02d",$mem_no);
  my $cmd = "fefe02801d08${mem_no}";
  my ($call,$slash) = split('/',$full_call);
  $call = sprintf("%-8.8s",$call);
  $slash = sprintf("%-4.4s",$slash);
  $call = unpack("H*",$call);
  $slash = unpack("H*",$slash);
  $cmd .= $call . $slash . "fd";
  if( &send2($cmd) ) {
    warn "Invalid MyCall data for #$mem_no\n";
  }
  else {
    if ($sel eq "S") {
      # This is the one to select.
      &send2("fefe02801d02${mem_no}fd");
    }
  }
}

#
# Get the number of the selected txt message
#
sub get_sel_txmsg {
  my $cmd = "fefe02801de1fd";
  &send($cmd);
  my $msg = &recv("fefe80021de1");
  my $sel_msg = substr($msg,12,2);
  $sel_msg += 0;
  return $sel_msg;
}

#
# Get the contents of the specified text message memory.
#
sub get_txmsg {
  my $msg_no = shift;
  $msg_no = sprintf("%02d",$msg_no);
  my $cmd = "fefe02801de0${msg_no}fd";
  &send($cmd);
  my $msg = &recv("fefe80021de0${msg_no}");
  my $txmsg = pack("H*",substr($msg,14,40));
  return($txmsg);
}

#
# Write a 'TxMsg' memory to the radio.
#
sub put_txmsg {
  my $msg_no = shift;
  my $sel = shift;
  my $txmsg = shift;
  $msg_no = sprintf("%02d",$msg_no);
  $txmsg = sprintf("%-20.20s",$txmsg);
  $txmsg = unpack("H*",$txmsg);
  my $cmd = "fefe02801de0${msg_no}${txmsg}fd";
  if( &send2($cmd) ) {
    warn "Invalid TxtMsg #$msg_no\n";
  }
  else {
    if ($sel eq "S") {
      # This is the one to select.
      &send2("fefe02801de1${msg_no}fd");
    }
  }
}

# get the name for a specified channel bank
sub get_bank_name {
  my $band = shift; # A or B
  my $bank = shift;
  my $nband = $band eq "A" ? "01" : "02";
  $bank = unpack("H2",$bank);
  my $cmd = "fefe${nband}801a0b${bank}fd";
  &send($cmd);
  my $msg = &recv("fefe80${nband}1a0b${bank}");
  my $name = pack("H*",substr($msg,14,16));
  return $name;
}

# Write the specified bank name to the radio.
sub put_bank_name {
  my $band = shift;
  my $bank = shift;
  my $name = shift;
  my $nband = $band eq "A" ? "01" : "02";
  $bank = unpack("H2",$bank);
  my $xname = sprintf("%-8.8s",$name);
  $xname = unpack("H*",$xname);
  my $cmd = "fefe${nband}801a0b${bank}${xname}fd";
  if (&send2($cmd)) {
    warn "Unable to set Bank name: $band $bank $name\n";
  }
}

#
# Get bank link data for specified band.
# Returns a string of bank letters linked.
#
sub get_bank_link {
  my $band = shift;
  my $nband = $band eq "A" ? "01" : "02";
  # Bitmap description
  # 87654321 87654321 87654321 87654321
  #  ABCDEFG  HIJKLMN  OPQRSTU  VWXYZ
  my $cmd = "fefe${nband}801a0532fd";
  &send($cmd);
  my $msg = &recv("fefe80${nband}1a0532");
  my $bldata = substr($msg,14,8);
  $bldata = pack("H*",$bldata);
  $bldata = unpack("B*",$bldata);
  my $linkmap = "0ABCDEFG0HIJKLMN0OPQRSTU0VWXYZ00";
  my $banklink = "";
  for (my $i=0;$i<length($linkmap);$i++) {
    my $x = substr($linkmap,$i,1);
    next if ($x eq "0");
    if ( substr($bldata,$i,1) eq "1") {
      $banklink .= $x;
    }
    else {
      $banklink .= "_";
    }
  }
  return $banklink;
}

#
# Send bank link data to radio for the specified band.
# parameters: Band, String of Bank letters linked.
#
sub put_bank_link {
  my $band = shift;
  my $banklink = shift;
  my $nband = $band eq "A" ? "01" : "02";
  my $linkmap = "0ABCDEFG0HIJKLMN0OPQRSTU0VWXYZ00";
  my $bldata = "";

  for (my $i=0; $i<length($linkmap); $i++) {
    my $x = substr($linkmap,$i,1);
    if ($x eq "0") {
      $bldata .= "0";
    }
    else {
      $bldata .= index($banklink,$x) >= 0 ? "1" : "0";
    }
  }
  $bldata = pack("B*",$bldata);
  $bldata = unpack("H*",$bldata);
  my $cmd = "fefe${nband}801a0532${bldata}fd";
  if (&send2($cmd)) {
    warn "Unable to set Bank links for band $band.\n";
  }
}

#
# Get the contents of a Rx Call record
#
sub get_rxlog {
  my $msg_no = shift;
  $msg_no = sprintf("%02x",$msg_no);
  my $cmd = "fefe02801d04${msg_no}fd";
  &send($cmd);
  my $msg = &recv("fefe80021d04${msg_no}");
  my $rx_stat = substr($msg,14,2);
  my $rx_rpt2 = pack("H*",substr($msg,16,16));
  my $rx_rpt1 = pack("H*",substr($msg,32,16));
  my $rx_called = pack("H*",substr($msg,48,16));
  my $rx_caller = pack("H*",substr($msg,64,16));
  my $rx_slash = pack("H*",substr($msg,80,8));
  my @rx_values = ($rx_stat,$rx_rpt2,$rx_rpt1,$rx_called,$rx_caller,$rx_slash);
  return \@rx_values;
}

#
# Get the number of the selected DTMF memory
#
sub get_sel_dtmf {
  my $cmd = "fefe01801a07fd";
  &send($cmd);
  my $msg = &recv("fefe80011a07");
  my $sel_dtmf = substr($msg,12,2);
  $sel_dtmf += 0;
  return $sel_dtmf;
}

#
# Get a DTMF Memory
#
sub get_dtmf {
  my $m_no = shift;
  $m_no = sprintf("%02x",$m_no);
  my $cmd = "fefe01801a06${m_no}fd";
  &send($cmd);
  my $msg = &recv("fefe80011a06${m_no}");
  my $number = pack("H*",substr($msg,14,32));
  my $ff = pack("H*","ff");
  $number =~ s/$ff//g;  #padded with 0xff characters.
  return $number;
}

# Write a DTMF memory
sub put_dtmf {
  my $m_no = shift;
  my $sel = shift;
  my $number = shift;
  $number = sprintf("%-16.16s",$number);
  my $xm_no = sprintf("%02x",$m_no);
  my $number = unpack("H*",$number);
  my $cmd = "fefe01801a06${xm_no}${number}fd";
  if (&send2($cmd)) {
    warn "Unable to set DTMF memory $m_no.\n";
  }
  else {
    if ($sel eq "S") {
      # This is the one to select.
      &send2("fefe01801a07${xm_no}fd");
    }
  }
}

#
# Get a Your Call memory
#
sub get_urcall {
  my $m_no = shift;
  $m_no = sprintf("%02x",$m_no);
  my $cmd = "fefe02801d06${m_no}fd";
  &send($cmd);
  my $msg = &recv("fefe80021d06${m_no}");
  my $urcall = pack("H*",substr($msg,14,16));
  return $urcall;
}

#
# Put a Your Call memory
#
sub put_urcall {
  my $m_no = shift;
  my $call = shift;
  my $xm_no = sprintf("%02x",$m_no);
  $call = sprintf("%-8.8s",$call);
  $call = unpack("H*",$call);
  my $cmd = "fefe02801d06${xm_no}${call}fd";
  if (&send2($cmd)) {
    warn "Unable to set UrCall memory $m_no.\n";
  }
}

sub put_tv {
  my($vband,$m_no,$chan_id,$freq,$mode,$skip) = @_;
  my $xm_no = sprintf("%02d",$m_no);
  # frequency - 5 bytes BCD
  my $xfreq = "";
  if ($freq == -1 || $freq == 0) {
    # clear this memory channel
    my $cmd = "fefe01801a0003${xm_no}fffd";
    print STDERR "$vband $m_no  \r";
    if( &send2($cmd)) {
      warn "\nInvalid data for $vband $m_no\n";
    }
    next;
  }
  else {
    $freq *= 1000000;
    $freq = int($freq + 0.1);
    $freq = sprintf("%010d",$freq);
    $xfreq = substr($freq,8,2) . substr($freq,6,2) . substr($freq,4,2) .
                substr($freq,2,2) . substr($freq,0,2);

  }
  # Mode (AM/FM...)
  my $xmode = 2; #default to wFM
  for (my $i=0; $i<=$#mod_mode; $i++) {
    if ($mod_mode[$i] eq $mode) {
      $xmode = $i;
      last;
    }
  }
  $xmode <<= 1;
  if ($skip =~ /Skip/i) {
    $xmode |= 1;
  }

  $xmode = sprintf("%02x",$xmode);

  my $xchan = unpack("H*",sprintf("%-2.2s",$chan_id));
  my $cmd = "fefe01801a0003${xm_no}${xfreq}${xmode}${xchan}fd";
  if (&send2($cmd)) {
    warn "Unable to set TV memory $m_no.\n";
  }
}

#
# Get a Rptr Call memory
#
sub get_rpcall {
  my $m_no = shift;
  $m_no = sprintf("%02x",$m_no);
  my $cmd = "fefe02801d07${m_no}fd";
  &send($cmd);
  my $msg = &recv("fefe80021d07${m_no}");
  my $rpcall = pack("H*",substr($msg,14,16));
  return $rpcall;
}

#
# Put a Your Call memory
#
sub put_rpcall {
  my $m_no = shift;
  my $call = shift;
  my $xm_no = sprintf("%02x",$m_no);
  $call = sprintf("%-8.8s",$call);
  $call = unpack("H*",$call);
  my $cmd = "fefe02801d07${xm_no}${call}fd";
  if (&send2($cmd)) {
    warn "Unable to set RpCall memory $m_no.\n";
  }
}

# get memory channel data
sub get_data {
  my $msg;
  my $rcnt = 0;
  open(CSV,">$csvfile") || die "Can't open output file: $csvfile\n";

  &get_ID; # make sure we can talk to radio and it's the right model
  $last_band = "0";


  # Output csv heading line
  if( $mgopt{bandA} || $mgopt{bandB}) {
    print CSV
      "\"#Band\"," .
      "\"ChNo\"," .
      "\"Frequency\"," .
      "\"Duplex\"," .
      "\"Offset\"," .
      "\"TS\"," .
      "\"Mode\"," .
      "\"Name\"," .
      "\"Skip\"," .
      "\"Tone\"," .
      "\"Rpt tone\"," .
      "\"CTCSS\"," .
      "\"DTCS\"," .
      "\"TR Pol\"," .
      "\"Bank\"," .
      "\"CH\"," .
      "\"DSQL\"," .
      "\"DS Code\"," .
      "\"YourCall\"," .
      "\"Rptr 1  \"," .
      "\"Rptr 2  \"," .
      "\"Train Freq\"" .
      "\n";
    }
  #RS-91 display order:
  #CH Freq DUP Offset TS Mode Name Skip Tone RptTone TSQL DTCS Polarity DSQL Code Your RPT1 Rpt2 Group CH
  for (my $band = 1; $band <= 2; $band++) {
    next if ($band == 1 && !$mgopt{bandA});
    next if ($band == 2 && !$mgopt{bandB});
    my $xband = sprintf("%02d",$band);
    my $vband = $band == 1 ? "A" : "B";
    my $seq_memch = -1;
    for (my $memch = 0; $memch < 850; $memch++) {
      last if ($band == 2 && $memch >= 450);
      my $xmemch = sprintf("%04d",$memch);
      my $fetch_cmd = "fefe${xband}801a0001${xmemch}fd";

      # Ask radio for what is in this memory location
      &send($fetch_cmd);
      $msg = &recv("fefe80${xband}1a0001");

      # decode the output
      &show_status($vband,"$vband $xmemch");
      my $aref = &decode_memory($band,1,$msg);

      if ($#$aref < 0) {
        next; #nothing in this memory location
      }

      if ($seq_memch == $$aref[1] ) {
        $$aref[1] = "* $$aref[1]";
        $seq_memch++;
      }
      else {
        $seq_memch = $$aref[1] + 1;
      }

      &write_csv($aref);
      $rcnt++;
    }

    # call channels
    for (my $callch=0; $callch < 2 ; $callch++) {
      my $xmemch = sprintf("%04d",$callch);
      my $fetch_cmd = "fefe${xband}801a0002${xmemch}fd";

      # Ask radio for what is in this memory location
      &send($fetch_cmd);
      $msg = &recv("fefe80${xband}1a0002");

      # decode the output
      #&show_status("$vband,"Call-$callch);
      my $aref = &decode_memory($band,2,$msg);

      if ($#$aref < 0) {
        next; #nothing in this memory location
      }

      &write_csv($aref);
      $rcnt++;
    }

    # test 05
    #{
    #  #my $xmemch = sprintf("%02d",$i);
    #  my $xmemch = "";
    #  my $fetch_cmd = "fefe${xband}801a000005fd";
    #
    #  # Ask radio for what is in this memory location
    #  &send($fetch_cmd);
    #  $msg = &recv("fefe80${xband}1a000005");
    #  $msg = substr($msg,0,13) . "00" . substr($msg,14);
    #  $msg = substr($msg,0,63) . "2020202020202020" . substr($msg,64);
    #
    #  # decode the output
    #  #&show_status("$vband,"Call-$callch);
    #  my $aref = &decode_memory($band,2,$msg);
    #
    #  if ($#$aref >= 0) {
    #    &write_csv($aref);
    #    $rcnt++;
    #  }
    #
    #}
    # test 06
    #{
    #  #my $xmemch = sprintf("%02d",$i);
    #  #my $xmemch = "";
    #  my $fetch_cmd = "fefe${xband}801a000006fd";
    #
    #  # Ask radio for what is in this memory location
    #  &send($fetch_cmd);
    #  $msg = &recv("fefe80${xband}1a000006");
    #  $msg = substr($msg,0,13) . "00" . substr($msg,14);
    #  $msg = substr($msg,0,63) . "2020202020202020" . substr($msg,64);
    #
    #  # decode the output
    #  #&show_status("$vband,"Call-$callch);
    #  my $aref = &decode_memory($band,2,$msg);
    #
    #  if ($#$aref >= 0) {
    #    &write_csv($aref);
    #    $rcnt++;
    #  }
    #}
    ## test 07
    #{
    #  #my $xmemch = sprintf("%02d",$i);
    #  my $xmemch = "";
    #  my $fetch_cmd = "fefe${xband}801a000007fd";
    #
    #  # Ask radio for what is in this memory location
    #  &send($fetch_cmd);
    #  $msg = &recv("fefe80${xband}1a000007");
    #  $msg = substr($msg,0,13) . "00" . substr($msg,14);
    #  $msg = substr($msg,0,63) . "2020202020202020" . substr($msg,64);
    #
    #  # decode the output
    #  #&show_status("$vband,"Call-$callch);
    #  my $aref = &decode_memory($band,2,$msg);
    #
    #  if ($#$aref >= 0) {
    #    &write_csv($aref);
    #    $rcnt++;
    #  }
    #}
  }
  # Get mycall data
  if ($mgopt{mycall}) {
    print CSV "\"#MyCall\",\"#\",\"Select\",\"CallSign/info\"\n";
    my $sel_call = &get_sel_mycall;
    for (my $i=1; $i<=6; $i++) {
      show_status("MyCall","Mycall $i");
      my $call = &get_mycall($i);
      print CSV "\"MyCall\",$i,";
      if ($i == $sel_call) {
        print CSV "\"S\",";
      }
      else {
        print CSV "\" \",";
      }
      print CSV "\"$call\"\n";
      $rcnt++;
    }
  }

  # Get your call data
  if ( $mgopt{urcall} ) {
    print CSV "\"#UrCall\",\"#\",\"CallSign\"\n";
    for (my $i=1; $i<=60; $i++) {
      show_status("UrCall","UrCall $i");
      my $call = &get_urcall($i);
      print CSV "\"UrCall\",$i,\"$call\"\n";
      $rcnt++;
    }
  }

  # Get Rptr call data
  if ( $mgopt{rpcall} ) {
    print CSV "\"#RpCall\",\"#\",\"CallSign\"\n";
    for (my $i=1; $i<=60; $i++) {
      show_status("RpCall","RpCall $i");
      my $call = &get_rpcall($i);
      print CSV "\"RpCall\",$i,\"$call\"\n";
      $rcnt++;
    }
  }

  # get TX message memory
  if ($mgopt{txmsg}) {
    print CSV "\"#TxMsg\",\"#\",\"Select\",\"Message\"\n";
    my $sel_txmsg = &get_sel_txmsg;
    for (my $i=1; $i<=5; $i++) {
      show_status("TxMsg","TxMsg $i");
      my $txmsg = &get_txmsg($i);
      $txmsg =~ s/"/\\"/g;  # escape quotes
      print CSV "\"TxMsg\",$i,";
      if ($i == $sel_txmsg) {
        print CSV "\"S\",";
      }
      else {
        print CSV "\" \",";
      }
      print CSV "\"$txmsg\"\n";
      $rcnt++;
    }
  }

  # get DTMF memories
  if ($mgopt{dtmf}) {
    print CSV "\"#DTMF\",\"#\",\"Select\",\"Number\"\n";
    my $sel_dtmf = &get_sel_dtmf;
    for( my $i=1; $i<=10; $i++) {
      show_status("DTMF","DTMF $i");
      my $pnumber = &get_dtmf($i);
      print CSV "\"DTMF\",$i,";
      if ($i == $sel_dtmf) {
        print CSV "\"S\",";
      }
      else {
        print CSV "\" \",";
      }
      print CSV "\"$pnumber\"\n";
      $rcnt++;
    }
  }

  # get bank names
  if ($mgopt{banknm}) {
    print CSV "\"#BankNm\",\"Band\",\"Bank\",\"Name\"\n";
    my $bank_list = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    for my $rband ("A","B") {
      for (my $i=0; $i < length($bank_list); $i++) {
        my $bank = substr($bank_list,$i,1);
        my $bname = &get_bank_name($rband,$bank);
        $bname =~ s/"/\\"/g; # escape any quotes
        print CSV "\"BankNm\",\"$rband\",\"$bank\",\"$bname\"\n";
        &show_status("BankNm","BankNm $rband $bank");
        $rcnt++;
      }
    }
  }

  # get bank link data for the band.
  if ($mgopt{bankln}) {
    for my $rband ("A","B") {
      my $bldata = &get_bank_link($rband);
      print CSV "\"BankLn\",\"$rband\",\"$bldata\"\n";
      &show_status("BankLn","BankLn $rband");
      $rcnt++;
    }
  }

  # get RX call records
  if ($mgopt{rxlog}) {
    print CSV "\"#RxLog\",\"#\",\"Caller\",\"/\",\"Called\"," .
              "\"Rx RPT1\",\"Rx RPT2\",\"Status\"\n";
    for( my $i=0; $i<20; $i++) {
      show_status("RxLog","RxLog $i");
      my $rxlog = &get_rxlog($i);
      print CSV "\"RxLog\",$i," .
        "\"" . $$rxlog[4] . "\"," .
        "\"" . $$rxlog[5] . "\"," .
        "\"" . $$rxlog[3] . "\"," .
        "\"" . $$rxlog[2] . "\"," .
        "\"" . $$rxlog[1] . "\"," .
        "\"" . $$rxlog[0] . "\"\n";
      $rcnt++;
    }
  }

  # get TV channel memories
  if ($mgopt{tv}) {
    print CSV "\"#TVchan\",\"#\",\"ID\",\"Frequency\",\"Mode\"," .
              "\"Skip\"\n";
    # These only exist in band "A"
    for (my $memch = 0; $memch < 70; $memch++) {
      my $xmemch = sprintf("%02d",$memch);
      my $fetch_cmd = "fefe01801a0003${xmemch}fd";
      # Ask radio for what is in this memory location
      &send($fetch_cmd);
      $msg = &recv("fefe80011a0003");

      # decode the output
      &show_status("TV","TV $memch");

      my $rmemch = substr($msg,14,2);
      $rmemch += 0;

      $rmemch = sprintf("%02d",$rmemch);

      my $freq = substr($msg,24,2) . substr($msg,22,2) . substr($msg,20,2) .
             substr($msg,18,2) . substr($msg,16,2);
      $freq /= 1000000;
      $freq = sprintf("%8.6f",$freq);

      my $mode = hex(substr($msg,26,2));
      my $skip = $mode & 1 ? "Skip" : "";
      $mode >>= 1;
      $mode = $mod_mode[$mode];

      my $chan_id = pack("H*",substr($msg,28,4));

      print CSV "\"TV\",\"$rmemch\",\"$chan_id\",\"$freq\",\"$mode\",\"$skip\"\n";
      $rcnt++;
    }
  }

  close(CSV);
  print STDERR "\nWrote $rcnt memory channels to file: $csvfile\n";
}

sub decode_memory {
  my $band = shift; # 1=A 2=B
  my $mtype = shift; # 1=Memory/PS 2=Call
  my $msg = shift;

  my @memdata;  # return reference to this array

  # decode and output
  # 000000000011 11 1111 1122222222 2233333333 3344 4444
  # 012345678901 23 4567 8901234567 8901234567 8901 2345
  # fefe80011a00 01 0000 0000044701 0000600000 1230 1230
  #
  # 4444555555 55 55 66 66 6666667777777777
  # 6789012345 67 89 01 23 4567890123456789
  # 0023228004 0e 00 41 00 4b37444156202020fd
  #
  # Offset 13: 01=Memory channel 02=Call channel 03=TV Channel
  #            00=VFO memory - Note this data has a different format!
  #            04= Bank channel (followed by bank letter and channel number)
  # Offset 14-17: Memory channel number BCD (0-849, 800-849 are PS memories)
  # 
  #
  # Offset 18-27: frequency LSB-MSB BCD if 'ff' in 18-19 then this channel is unused.
  #        28-37: Offset LSB-MSB  BCD
  #        38-41: Repeater tone * 10 BCD
  #        42-45: CTCSS tone * 10 BCD
  #        46-49: DTCS tone BCD
  #        50-53: '2280' Train Frequency
  #           54: Mode AM=3 DV=4 FM=0 FM-N=1 (Assmunig wFM=2)
  #           55: Tuning step 200=d 125=c 100=b 50=a 30=9
  #               25=8 20=7 15=6 12.5=5 10=4 6.25=1 5=0  (Assuming 9=3 7.5=2)
  #        56-57: Bits 1-2  Duplex +=2 -=1 off=0
  #               Bits 3-5  Tone  TONE=4 TSQL=c off=0 DTCS=14
  #        58-59: Bits 1-2 Skip 0=off 1=skip  2=Pskip
  #               Bits 3-4 Polarity 0=NN 4=NR 8=RN c=RR
  #        60-61: Bank Group (A-Z)
  #        62-63: Bank channel # 00-99 BCD
  #        64-79: Channel Name
  #
  #        ### BAND B ONLY ###
  #        80-81: DSQL DSQL=02 OFF=00 CSQL=04 TODO: DBEEP=01 CBEEP=03
  #        82-83: DSQL code 00-99 BCD
  #        84-99: RPT2
  #      100-115: RPT1
  #      116-131: Your Call
  #
  #   Possibilites for 50-53:
  #       Auto-reply, DV Data TX, Digital Monitor, RX Call sign auto write
  #       Power, Rptr call sign auto write, GPS mode
  #       (None of these seem too likely)
  #
  my $vband = substr($msg,6,2);
  $vband = $vband eq "01" ? "A" : "B";

  my $mem_type = substr($msg,12,2);

  my $rmemch = substr($msg,14,4);
  $rmemch += 0;
  my $last_mem = 800;
  $last_mem = 400 if( $vband eq "B");

  # handle special names of Programmed scan memory
  if ($mem_type eq "01") {
    if ($rmemch >= $last_mem && $rmemch < ($last_mem + 50)) {
      my $psnum = 0;
      my $psmem = $rmemch - $last_mem;
      $psnum = $psmem / 2;
      my $ab = $psmem % 2 == 1 ? "b" : "a";
      $rmemch = sprintf("PS-%02d-%s",$psnum,$ab);
    }
  }
  # Call channel memory - Format memory name
  elsif ($mem_type eq "02") {
    $rmemch = sprintf("Call-%d",$rmemch);
  }
  elsif ($mem_type eq "00") {
    $rmemch = sprintf("V-%02d",$rmemch);
  }

  if ( substr($msg,18,2) eq "ff" ) {
    # nothing in here, it's empty
    return \@memdata;
  }
  my $freq = substr($msg,26,2) . substr($msg,24,2) . substr($msg,22,2) .
             substr($msg,20,2) . substr($msg,18,2);
  $freq /= 1000000;
  $freq = sprintf("%8.6f",$freq);

  my $offset = substr($msg,36,2) . substr($msg,34,2) . substr($msg,32,2) . 
               substr($msg,30,2) . substr($msg,28,2);
  $offset /= 1000000;
  $offset = sprintf("%8.6f",$offset);

  my $tone = substr($msg,38,4);
  $tone /= 10;
  $tone = sprintf("%5.1f",$tone);
  my $ctcss = substr($msg,42,4);
  $ctcss /= 10;
  $ctcss = sprintf("%5.1f",$ctcss);

  my $dtcs = substr($msg,46,4);
  $dtcs += 0;

  my $train_freq = substr($msg,50,4);

  my $mode = substr($msg,54,1);
  $mode = $mod_mode[$mode];

  my $tstep = substr($msg,55,1);
  $tstep = hex($tstep);
  $tstep = $tstep_code[$tstep];

  my $dup = hex(substr($msg,57,1)) & 0x03;
  $dup = $dup_code[$dup];

  my $tone_type = hex(substr($msg,56,2));
  $tone_type &= 0x1c;
  $tone_type = $tone_code{$tone_type};

  my $skip = hex(substr($msg,58,2)) & 0x03;
  $skip = $skip_code[$skip];

  my $polarity = hex(substr($msg,58,2)) & 0x0c;
  $polarity = $pol_code{$polarity};

  my $bank = substr($msg,60,2);
  if ($bank eq "00") {
    # No bank assigment for this memory channel
    $bank = "";
  }
  else {
    $bank = pack("H*",$bank);
  }
  my $bchan = substr($msg,62,2);
  $bchan += 0;

  my $chan_name = pack("H*",substr($msg,64,16));

  my $urcall = "";
  my $rpt1 = "";
  my $rpt2 = "";
  my $dsql = "";
  my $dsql_code = "";
  if( $band == 2 ) {
    $dsql = hex(substr($msg,80,2)) & 0x07;
    $dsql = $dsql_type{$dsql};

    $dsql_code = substr($msg,82,2);

    $rpt2 = pack("H*",substr($msg,84,16));
    $rpt1 = pack("H*",substr($msg,100,16));
    $urcall = pack("H*",substr($msg,116,16));
  }

  @memdata = (
        $vband,
        $rmemch,
        $freq,
        $dup,
        $offset,
        $tstep,
        $mode,
        $chan_name,
        $skip,
        $tone_type,
        $tone,
        $ctcss,
        $dtcs,
        $polarity,
        $bank,
        $bchan,
        $dsql,
        $dsql_code,
        $urcall,
        $rpt1,
        $rpt2,
        $train_freq
  );
  return \@memdata;
}

sub write_csv {
  my $aref = shift;
  my (
    $vband,
    $rmemch,
    $freq,
    $dup,
    $offset,
    $tstep,
    $mode,
    $chan_name,
    $skip,
    $tone_type,
    $tone,
    $ctcss,
    $dtcs,
    $polarity,
    $bank,
    $bchan,
    $dsql,
    $dsql_code,
    $urcall,
    $rpt1,
    $rpt2,
    $train_freq
  ) = @$aref;

  # escape any quotes
  $chan_name =~ s/"/\\"/g; # escape any quotes
  print CSV
    "\"$vband\"," .
    "$rmemch," .
    "$freq," .
    "\"$dup\"," .
    "$offset," .
    "$tstep," .
    "\"$mode\"," .
    "\"$chan_name\"," .
    "\"$skip\"," .
    "\"$tone_type\"," .
    "$tone," .
    "$ctcss," .
    "$dtcs," .
    "\"$polarity\"," .
    "\"$bank\"," .
    "$bchan," .
    "\"$dsql\"," .
    "$dsql_code," .
    "\"$urcall\"," .
    "\"$rpt1\"," .
    "\"$rpt2\"," .
    "$train_freq," .
    "\n";
}

sub show_status {
  my $band = shift;
  my $info = shift;
  if ($last_band ne "0"  && $last_band ne $band) {
    print STDERR "\n";
  }
  $last_band = $band;
  print STDERR "$info   \r";
}

sub put_data {
  # writing memory
  # computer sends something similar to:
  #  fe fe 01 80 1a 00 01 00 00 00 00 04 47 01 00 00 60 00 00 12
  #  30 12 30 00 23 22 80 04 0e 00 41 00 4b 37 44 41 56 20 20 20 fd
  #
  # radio responds:
  #  fe fe 80 01 fb fd  - OK
  #  fe fe 80 01 fa fd  - NG (not valid/incorrect setting)
  my $wcnt = 0;
  my $msg;
  open(CSV,$csvfile) || die "Can't open input file: $csvfile\n";
  $last_band = "0";

  &get_ID; #Check radio and connection.

  my $banklink = "";
  my $seq_memch = -1; # for auto sequencing
  while( <CSV> ) {
    chomp;
    next if (length($_) == 0); # skip blank lines
    next if (/^"?#/);  # skip lines that start with '"#'
    next if (/^#/);  # skip lines that start with '#'

    my @csvdata = quotewords(",", 0, $_);    # split(",",$_);

    if ( $csvdata[0] =~ /^TV/ ) {
        next if ( !$mgopt{tv} );
        my ($vband,$m_no,$chan_id,$freq,$mode,$skip) = @csvdata[0..5];
        &put_tv($vband,$m_no,$chan_id,$freq,$mode,$skip);
        &show_status($vband,"TV $m_no");
        $wcnt++;
        next;
    }

    if ( $csvdata[0] =~ /^RxLog/ ) {
      next; # Don't load this.
    }

    if ( $csvdata[0] =~ /^DTMF/ ) {
      next if ( !$mgopt{dtmf} );
      my ($vband,$m_no,$sel,$number) = @csvdata[0..3];
      &put_dtmf($m_no,$sel,$number);
      &show_status($vband,"DTMF $m_no");
      $wcnt++;
      next;
    }

    if ( $csvdata[0] =~ /^MyCall/ ) {
      next if ( !$mgopt{mycall} );
      my ($vband,$mc_no,$sel,$call) = @csvdata[0..3];
      &put_mycall($mc_no,$sel,$call);
      &show_status($vband,"MyCall $mc_no");
      $wcnt++;
      next;
    }

    if ( $csvdata[0] =~ /^UrCall/ ) {
      next if ( !$mgopt{urcall} );
      my ($vband,$m_no,$call) = @csvdata[0..2];
      &put_urcall($m_no,$call);
      &show_status($vband,"UrCall $m_no");
      $wcnt++;
      next;
    }

    if ( $csvdata[0] =~ /^RpCall/ ) {
      next if ( !$mgopt{rpcall} );
      my ($vband,$m_no,$call) = @csvdata[0..2];
      &put_rpcall($m_no,$call);
      &show_status($vband,"RpCall $m_no");
      $wcnt++;
      next;
    }

    if ( $csvdata[0] =~ /^TxMsg/ ){
      next if ( !$mgopt{txmsg} );
      my ($vband,$txmsg_no,$sel,$txmsg) = @csvdata[0..3];
      &put_txmsg($txmsg_no,$sel,$txmsg);
      &show_status($vband,"TxMsg $txmsg_no");
      $wcnt++;
      next;
    }
    if ( $csvdata[0] =~ /^BankNm/ ) {
      next if ( !$mgopt{banknm} );
      my ($vband,$xband,$xbank,$xname) = @csvdata[0..3];
      &put_bank_name($xband,$xbank,$xname);
      &show_status($vband,"$vband $xband $xbank");
      $wcnt++;
      next;
    }
    if ( $csvdata[0] =~ /^BankLn/ ) {
      next if ( !$mgopt{bankln} );
      my ($vband,$xband,$bldata) = @csvdata[0..2];
      &put_bank_link($xband,$bldata);
      &show_status($vband,"$vband $xband");
      $wcnt++;
      next;
    }

    my(
        $vband,
        $rmemch,
        $freq,
        $dup,
        $offset,
        $tstep,
        $mode,
        $chan_name,
        $skip,
        $tone_type,
        $tone,
        $ctcss,
        $dtcs,
        $polarity,
        $bank,
        $bchan,
        $dsql,
        $dsql_code,
        $urcall,
        $rpt1,
        $rpt2,
        $train_freq
      ) = @csvdata;


    next if ($vband eq "A" and !$mgopt{bandA});
    next if ($vband eq "B" and !$mgopt{bandB});

    # build initial command message to program a radio channel
    my $xband = "01"; # Band A
    $xband = "02" if ($vband eq "B");
    $msg = "fefe${xband}801a00";

    # Channel type: 01=Memory 02=Call 03=TV
    my $xmemch;
    my $xchan_type = "01"; #default to regular memory channel
    if ($rmemch =~ /^\d+$/) {
      # standard channel number, set seq_memch
      $seq_memch = $rmemch;
    }
    elsif ($rmemch =~ /^\*/) { # Use channel of previous + 1
      $rmemch = ++$seq_memch;
    }

    if ($rmemch =~ /^PS/) {
      $xchan_type = "01";
      # TODO: adjust memory channel number
      my @psinfo = split('-',$rmemch);
      my $last_mem = 800;
      $last_mem = 400 if ($vband eq "B");
      $xmemch = $last_mem + $psinfo[1] * 2;
      if ($psinfo[2] eq "b") {
        $xmemch++;
      }
    }
    elsif ($rmemch =~ /^Call/) {
      $xchan_type = "02";
      my @info = split('-',$rmemch);
      $xmemch = $info[1];
    }
    else {
      $xmemch = $rmemch;
    }
    $xmemch = sprintf("%04d",$xmemch);

    # memory channel
    $msg .= $xchan_type . $xmemch;

    # frequency - 5 bytes BCD
    if ($freq == -1 || $freq == 0) {
      # clear this memory channel
      $msg .= "fffd";
      print STDERR "$vband $rmemch  \r";
      if( &send2($msg)) {
        warn "\nInvalid data for $vband $rmemch\n";
      }
      else {
        $wcnt++;
      }
      next;
    }
    else {
      $freq *= 1000000;
      $freq = int($freq + 0.1);
      $freq = sprintf("%010d",$freq);
      my $xfreq = substr($freq,8,2) . substr($freq,6,2) . substr($freq,4,2) .
                  substr($freq,2,2) . substr($freq,0,2);
      $msg .= $xfreq;

      # Offset
      $offset *= 1000000;
      $offset = int($offset + 0.1);
      $offset = sprintf("%010d",$offset);
      my $xoffset = substr($offset,8,2) . substr($offset,6,2) . substr($offset,4,2) .
                  substr($offset,2,2) . substr($offset,0,2);
      $msg .= $xoffset;

      # Rptr Tone
      my $xtone = sprintf("%04d",int($tone * 10));
      $msg .= $xtone;

      # CTCSS
      my $xctcss = sprintf("%04d",int($ctcss * 10));
      $msg .= $xctcss;

      # DTCS
      my $xdtcs = sprintf("%04d",$dtcs);
      $msg .= $xdtcs;

      # train_freq '2280'
      #$msg .= "2280";
      # Fri Dec 21 15:46:33 MST 2007
      #  I finally found out what this it. It is called 'Train Freq'.
      #  It is a tone frequency used to indicate a clear/unused line/channel.
      #  I assume it used by trains in Japan.  There are 2 associated
      #  'Tone' settings for these: "train1" and "train2".
      #  Train1 enables the 2280hz tone detection (or whatever freq is set)
      #  Squelch is closed when the tone is present.
      #  Train2 if for an alternating tone signal. (?1500hz/2280hz?) but
      #  I assume it works like Train1.
      #  Values: 300-3000Hz in 10Hz steps.
      if ($train_freq < 300 || $train_freq > 3000) {
          warn "Invalid Train Freq: $train_freq\n";
          $train_freq = "2280";
      }
      $train_freq = sprintf("%04d",$train_freq);
      $msg .= $train_freq;

      # Mode (AM/FM...)
      my $xmode = "00"; #default to FM
      for (my $i=0; $i<=$#mod_mode; $i++) {
        if ($mod_mode[$i] eq $mode) {
          $xmode = $i;
          last;
        }
      }
      $xmode = sprintf("%x",$xmode);
      $msg .= $xmode;

      # Tuning Step
      my $xtstep = -1;
      for (my $i=0; $i<=$#tstep_code; $i++) {
        if ($tstep_code[$i] eq $tstep) {
          $xtstep = $i;
          last;
        }
      }
      if ($xtstep == -1) {
        warn "$vband $rmemch: Invalid tuning step: $tstep  Defaulting to 5\n";
        $xtstep = 0;
      }
      $xtstep = sprintf("%x",$xtstep);
      $msg .= $xtstep;

      # Duplex and Tone_type
      my $xdup = -1;
      my $xtone = -1;
      for (my $i=0; $i<=$#dup_code; $i++) {
        if ($dup_code[$i] eq $dup) {
          $xdup = $i;
          last;
        }
      }
      if ($xdup == -1) {
        warn "$vband $rmemch: Invalid Duplex setting: $dup  Defaulting to simplex\n";
        $xdup = 0;
      }
      for my $key (keys %tone_code) {
        if ($tone_code{$key} eq $tone_type) {
          $xtone = $key;
          last;
        }
      }
      if ($xtone == -1) {
        warn "$vband $rmemch: Invalid Tone setting: $tone_type  Defaulting to OFF\n";
        $xtone = 0;
      }
      my $xdt = $xdup | $xtone;
      $xdt = sprintf("%02x",$xdt);
      $msg .= $xdt;

      # Skip and Polarity
      my $xskip = -1;
      my $xpol = -1;
      for (my $i=0; $i<=$#skip_code; $i++) {
        if ($skip_code[$i] eq $skip) {
          $xskip = $i;
          last;
        }
      }
      if ($xskip == -1) {
        warn "$vband $rmemch: Invalid Skip setting: $skip  Defaulting to OFF\n";
        $xskip = 0;
      }
      for my $key (keys %pol_code) {
        if ($pol_code{$key} eq $polarity) {
          $xpol = $key;
          last;
        }
      }
      if ($xpol == -1) {
        warn "$vband $rmemch: Invalid Polarity setting: $skip  Defaulting to NN\n";
        $xpol = 0;
      }
      my $xsp = $xskip | $xpol;
      $xsp = sprintf("%02x",$xsp);
      $msg .= $xsp;

      # Bank Group (A-Z)
      my $bg = "";
      if(length($bank) > 0) {
        $bank =~ y/a-z/A-Z/;
        $bg = unpack("H2",$bank);
      }
      else {
        $bg = "00";
      }
      $msg .= $bg;

      # Bank Channel (00-99)
      my $xbch = sprintf("%02d",$bchan);
      $msg .= $xbch;

      # Channel Name
      $chan_name = sprintf("%-8.8s",$chan_name);
      my $xname = unpack("H*",$chan_name);
      $msg .= $xname;


      if ($vband eq "B") {
        # ### Band B Only ###
        # DSQL type
        my $xdsql = -1;
        for my $key (keys %dsql_type) {
          if ($dsql_type{$key} eq $dsql) {
            $xdsql = $key;
            last;
          }
        }
        if ($xdsql == -1) {
          warn "$vband $rmemch: Invalid DSQL setting: $dsql  Defaulting to OFF\n";
          $xdsql = 0;
        }
        $xdsql = sprintf("%02x",$xdsql);
        $msg .= $xdsql;

        # DSQL code
        my $xdsql_code = sprintf("%02d",$dsql_code);
        $msg .= $xdsql_code;

        # Rpt2
        $rpt2 = sprintf("%-8.8s",$rpt2);
        my $xrpt2 = unpack("H*",$rpt2);
        $msg .= $xrpt2;

        # Rpt1
        $rpt1 = sprintf("%-8.8s",$rpt1);
        my $xrpt1 = unpack("H*",$rpt1);
        $msg .= $xrpt1;

        # Your Call
        $urcall = sprintf("%-8.8s",$urcall);
        my $xurcall = unpack("H*",$urcall);
        $msg .= $xurcall;

      }
    }
    # Command terminator
    $msg .= "fd";

    #print "msg: $msg\n";
    &show_status($vband,"$vband $rmemch");
    # send command and wait for acknowledgement
    if( &send2($msg) ) {
      warn "\nInvalid data for $vband $rmemch\n";
    }
    else {
      $wcnt++;
    }
  }
  print "\nWrote $wcnt memory channels to the radio.\n";

}

#
# Send command to radio and wait for it to be acknowledged.
#
sub send2 {
  my $msg = shift;
  &send($msg);

  # wait for acknowledge
  my $stat = 1;
  while ( 1 ) {
    my $ack = &recv("fefe800");
    if ($ack =~ /^fefe800.fbfd/) {
      # OK
      $stat = 0;
      last;
    }
    elsif ($ack =~ /^fefe800.fafd/) {
      # No good
      if ($debug) {
        warn "msg: $msg\n";
      }
      last;
    }
    else {
      if ($debug) {
        warn "Unexpected ACK response: $ack (still waiting)\n";
      }
    }
  }
  return $stat;
}

sub send {
  my $msg = shift;
  print STDERR "s:$msg\n" if ($debug);
  $msg = pack("H*",$msg);
  my $rv = syswrite(RADIO,$msg);
  warn "Write error\n" if ($rv < 0);
}

sub recv {
  my $expected = shift;
  my $msg = "";
  while( 1 ) {
    $msg = "";
    my $c = "";
    print STDERR "r:" if ($debug);
    while ( 1 ) {
      my $rv = sysread(RADIO,$c,1);
      if ($rv == 0) {
        sleep(1);
        next;
      }
      my $hx = unpack("H*",$c);
      $msg .= $hx;
      print STDERR "$hx" if ($debug);
      last if ($hx eq "fd");
    }
    print STDERR "\n" if ($debug);
    if ($msg !~ /^$expected/) {
      if ($msg =~/^fefe80..fafd/ ) {
        warn "Data/command rejected by radio\n";
        last;
      }
      #print "Unsolicited response: $msg\n";
      # print "Expected: $expected\n";
    }
    else {
      #print "Found proper response: $msg\n";
      last;
    }
  }
  return $msg;
}

#######################################################
#
# $Log: cx91,v $
# Revision 1.9  2008-05-30 03:11:04  bruce
# Fix bug when uploading TV channels to radio.
#
# Revision 1.8  2008-04-17 03:36:22  bruce
# Added support for TV channels.
#
# Revision 1.7  2008-01-04 04:34:53  bruce
# Add support for squelch settings: train2 and msk.
# Add support for train frequency setting.
#
# Revision 1.6  2007-11-11 04:39:19  bruce
# Fix bug setting bank names on band A.
#
# Revision 1.5  2007-09-27 22:12:18  bruce
# Allow '*' to indicate the channel number is to be automatically assigned
# when loading.  The channel number column can contain a number, a '*'
# or "* number" where the number is the current channel number. As long
# as the column begins with '*' the channel number will be automatically
# assigned.
#
# Revision 1.4  2007-09-27 14:32:33  bruce
# Force '2280' unknown memory config settings. Also document valid
# values, even though what effect it has is still unknown.
#
# Revision 1.3  2007-09-26 04:46:11  bruce
# Fix usage message.
#
# Revision 1.2  2007-09-26 04:38:13  bruce
# Add support for additional memory types: DTMF UrCall, RptCall ...
#
# Revision 1.1.1.1  2007-09-20 15:01:27  bruce
# Initial CVS import
#
#
#######################################################
