#!/usr/bin/perl

# Simple chat program for DStar
#
# After seeing "http://nj6n.com/dstar/dstar_chat.html"
# I thought I could do better with a simple command line program.
# 
# Design Ideas:
#  User interface: 'KG7WI> '
#    Input text to send.
#    /set urcall CQCQCQCQ
#    /set mycall KG7WI
#    /set port /dev/ttyS0
#    /set speed 19200
#    /set rpt1 WA7GIE B
#    /set rpt2 WA7GIE C
#    /set memory "B7" | "135"
#    /set vfo 145.60  (DV mode should be set automatically)
#    /set duplex -|+|off
#    /set power H|L
#    /set DCODE "00"  # digital code
#
#    /load <settings file>
#    /save <settings file>  (Default .drc.rc)
#    /send <msg file>
#    /quit -or- /exit -or- /bye -or- /73
#
#    Show received call info
#    fe fe 80 02 1d 00 00 00 02 fd  D-star header flag read out
#                                   Top flag = 0 Bottom Flag = 02
#    fe fe 80 02 1d 04 00 02 <rpt2(8)> <rpt1(8)> <called(8)> <caller(8)> <stroke(4)> fd
#     --- I think the  00 02 is a status --  UR?
#
#    Input message to send.
#
#    Set radio "DV Data TX" [Auto]
#    <  fe fe 02 80 1d 18 01 fd      DV Auto TX set ON
#    >  fe fe 80 02 fb fd            Ack
#       fe fe 80 01 1c 00 02 fd      TX(PTT) Read out 02=TX 00=RX
#       fe fe 80 01 1a 04 0e 00 fd   ?
#       fe fe 80 01 1c 00 00 fd      TX(PTT) Read out
#       fe fe 80 02 15 01 01 fd      SQL Open
#       fe fe 80 02 15 02 02 55 fd   S-Meter 255
#       fe fe 80 02 15 02 00 00 fd   S-Meter 0
#       fe fe 80 02 15 01 00 fd      SQL Closed
#
#       
#
#    Send message
#      fe fe 02 80 1d c1 nn b1 b2 ... bn fd  # send message of nn bytes
#      fe fe 80 02 fb fd                     #ack
#
#      nn is hex representation of number (1f = 31 bytes)
#
#
#    Set radio  "DV Data TX" [PTT]
#    <  fe fe 02 80 1d 18 00 fd      DV Auto TX set Off
#    >  fe fe 80 02 fb fd            Ack
#
#    /set gpsrx on|off
#    /set msg[1-9] "text"
#    /msg[1-9]  # sends pre-set message
#    /info  # shows info about radio, frequency, settings...
#    /log <logfile>  # start saving a copy of chat session to <logfile>
#
#    Problem: Need to handle simultaneous input from two sources (radio and keyboard).
#    Solution 1: Use curses or windowing interface so there is an input area and
#                an output area, similar to an IM screen. An outgoing message is
#                displayed in the output area when it is sent.
#                Need scrollback of output area.
#    Solution 2: Don't prompt for input until a special key is hit. This way incoming
#                Messages are displayed normally.
#    Solution 3: Prompt for input, but if a message comes in and the input line
#                is empty, erase the input prompt, display the message, then 
#                re-display the input prompt.  Don't display incoming messages
#                while an outgoing message is being input.  Alternately could we
#                keep what has been typed so far, erase it, display the incoming
#                message and then re-display the outgoing message?  This could
#                make message editing a pain.  It would be nice to be able
#                to use GNU readline for input, so that we had fancy command
#                line editing as well as history.
#
#
#  Advanced features.
#     Send a 1 line private message:
#      For example: URCALL is set to CQCQCQ
#          I do: /msg K7VE <msg>  (or perhaps /privmsg)
#          This sets URCALL to K7VE for this message only, then sets it back
#          To 'CQCQCQ' when done.
#
#     ACK mode.
#       Each message sent (probably not to CQCQCQ) includes a
#       checksum.  When the message is received, a ACK/NAK response is
#       sent (to the originator) to indicate successfull reciept of
#       the message.  If the sender receives a NAK, the message is 
#       automatically re-sent (number of re-tries configurable).
#       If a message is not received correclty it is not shown 
#       (or perhaps shown only in a debug mode, or shown as a message
#       received with errors.)
#       /set ack ON|OFF
#       If supported, Ack mode is automatically detected and used on
#       a channel.
#
#    Other features from IM and/or IRC clients.
#      - emoticons
#      - /nick nickname setting
#      - /me substitution.
#      - /debug
#      - /say (done)
#      - /ping callsign
#      - /heard (lists heard stations) 
#        (done), but it would be nice to have an internal list
#        that also had timestamps.
#      - /privmsg
#      - /info
#

# TODO
#  1. Allow escaping of initial '/' via '\'
#     (not as critical as /send can be used as a workaround)
#  2. debug long messages. (Done.)
#     a. Are they getting sent correctly? Yes, however I need a recevier to verify.
#     b. Is there really a 32 byte limit? RS-91 seems to observe this limit.
#        If so, I need to send newlines to indicate end of a message line,
#        and perhaps buffer input messages until a newline is received
#        before displaying them. (Working OK as coded.)
#     c. (Fixed) Why don't I see all of the ACKs? 
#  3. Add set command
#     a. (done) /set mycall  (/set mycall [v|<n>] kg7wi/91ad, /set mycall [v|<n>] kg7wi)
#        Call sign uppercased, after slash is left as-is.
#        /set mycall <n> - use mycall memory <n> as current call.
#        This should change the prompt accordingly.
#        This is done, except for 'v' (VFO) setting mode.
#
#     b. (done via /set txcall) /set urcall
#     c. (done via /set txcall) /set rpt1
#     d. (done via /set txcall) /set rpt2
#
#     e. (done) /set power [l|h]
#     f. (done via /set tone) /set dsql [on|off|beep]
#     g. (done via /set tone) /set csql [on|off|beep]
#     h. (done) /set dsql [off|DSQL|CSQL [code]]
#     i. /set txmsg <n> <TX message contents>
#  4. Add log command and logging features
#     a. /log <filename>
#     b. /log close
#  5. Add show command
#     a. (done) /show rx  - show received callsign history
#     b. (done) /show mycall - show mycall settings
#     c. (done) /show tx - show TX callsign settings.
#     d. (done) /show power
#     e. (done) /show vfo
#     f. (done) /show dsql
#  6. Don't transmit if squelch is open (Receiving something)
#     This may require the implementation of a tx queue.
#     May also want a setting to control this behaviour.
#     There is a radio setting for this, and a command
#     implemented here: /set busylo [on|off]
#     Perhaps this should be run at startup.
#     (Warning: leaving 'busylo on' can be bad for autopatch use!)
#
#  7. May need to use '1de2' Message RX callsign readout.
#     Does it give the callsign of the current/just received message?
#     If so, then I can get this after a message, and then use it
#     to prefix the message text.
#     Note: '1de2' is the setting for whether or not a message is sent
#     during TX. The functionality I was looking for here does not
#     seem to be available.

#  8. May want to implement an RX message queue.  Incomming messages
#     from a given source would be concatenated together and displayed only
#     after some 'end-of-message' indicator (perhaps just '\n') is received.
#     May also display after some amount of elapsed time if no 'end' indicator
#     is received.
#     In order to do this I need to figure out how to associate a
#     call sign with a received message, and I will need to see some
#     incomming messages in order to figure this out.
#  9. (Done) Honor Xon/Xoff flow control.  The radio may send ^S/^Q flow control.
#     (0x11=Xon 0x13=Xoff)  I've actually seen these in the data coming
#     from the radio when I send a long message.
# 10. Settings that I may want to duplicate here from D*chat
#     a) (done) gps off  (gps mode off)
#     b) setting of comm port and speed. (Currently done with command line options)
#     c) Send periodic QST with QST Message and Time (i.e. 30 minutes)
#     d) RxBeep - beep on receive
#     e) quick send messages (7)
#     f) show gps [on/off] (what does this do?)
#     g) show qst [on/off]
#     h) tx/rx colorized text with customizable colors
#     i) dChat end each message with CR/LF, and prefixes each message with
#        timestamp and call ('11:01:49AM: KG7WI> ' is part of message sent.
#        Sending the timestamp is a config option that can be turned on/off.
#     g) sent text history than can be recalled/editied.
#     Some of the above imply that there is a 'config' file that settings
#     can be loaded from at startup and saved to (by command and possbily
#     automatically at exit).
#
# 11. Is the mic active when transmitting text? (No)
# 12. Do I want a way to program memories?
#     a) Write VFO to a specified memory
#     b) Copy memory to VFO
# 13. Get/set memory name. I now have a /show mem command that shows
#     all info about the current memory.
# 14. /set rxcs -> sets TX callsign info from last heard RX callsign info for
#     replying to a station. Will need to get last heard and then use it
#     to set txinfo
# 15. Handle various incoming status messages from the radio.
# 16. Learn how to associate a received message with the senders call sign.
# 17. I might as well implement these radio control functions for completeness
#     a) (done) /set att on/off
#     b) (done) /set sql <level>
#     c) (done) /set vol <level>
# 18. Use GNU readline for input so that there is history and editing.
# 19. Generalize radio commands/responses. Possibly implement a table/hash
#     that contains radio commands so that the codes can be referred to
#     by name, and the redundant part (fe fe ...) of all messages can be removed.
# 20. I resist this, but this program probably needs to use a GUI, with
#     separate input and output areas. This makes it easier to handle
#     input editing and interruptions (incomming messages)
#     This would also make it easier to manage radio settings.
# 21. Keep settings in a startup file so defaults can be used for
#     command line options and radio settings.  Radio settings should
#     be set restored when the program exits.
# 22. Change model to an event-based / threaded model. (threads for
#     monitoring radio input/output, user input (perhaps more).
# 23. GPS Messages come in as standard messages and look like this:
#     $GPGGA,162459.01,4044.5634,N,11236.5620,W,1,08,1.0,1403.0,M,-18.0,M,,*6F
#     $GPGGA,163622.01,4049.1115,N,11252.2685,W,1,07,2.0,1497.0,M,-18.4,M,,*62
#     $GPRMC,162459.01,A,4044.56E
#     $GPRMC,162501.01,A,4044.5767,N,11236.6086,W,68.94,289.8,190309,13.2,E,A*15
#
#     It's interesting that I got a lot of partial messages, I wonder if I'm
#     missing them, or It's because of the repeater or a week signal.
# 24. There is a 'data loss' indicator, A flashing "L" above the memory
#     channel indicator.  I assume there is also something sent on the
#     interface to indicate this. Actually, It might show where the EMR/BK
#     indicator is.
#

# I found the following descriptions of status:
#   UR?: Target station does not reply
#   UR*: linking repeater not found
# It looks like the following might be more accurate:
#   RPT?: linking repeater not found
# So, What are the following (if they exist)?
#   UR*
#  RPT*

use strict;
use Fcntl;
use Getopt::Long;

########################################################
# Command line arguments
#
my $debug = 0;

# Default port and baud rate
my $port = "/dev/ttyS0";
#my $port = "/tmp/interceptty";
my $baud = 38400;

my $help = 0;
########################################################

my $radio_open = 0;

# Saved power setting mode.  Power saver is forced to off at startup and then
# restored when the program exits.
my $psave_state = "00";

&GetOptions( "port=s" => \$port, "baud=i" => \$baud, "debug" => \$debug,
             "help" => \$help );

# Memory setting definitions
my @mod_mode = ("FM","FM-N","wFM","AM","DV");
my @tstep_code = (
    "5",
    "6.25",
    "7.5",
    "9",
    "10",
    "12.5",
    "15",
    "20",
    "25",
    "30",
    "50",
    "100",
    "125",
    "200"
  );

my @dup_code = ("OFF","Dup-","Dup+");

my %tone_code;
$tone_code{0x00} = "OFF";
$tone_code{0x04} = "TONE";
$tone_code{0x08} = "TSQLPB";
$tone_code{0x0c} = "TSQL";
$tone_code{0x10} = "DTCSPB";
$tone_code{0x14} = "DTCS";
$tone_code{0x18} = "TSQLR"; # Inversion tone squelch (Non-USA)
$tone_code{0x1c} = "TRAIN1"; # Mute if 'train freq' detected (empty-line squelch)
$tone_code{0x20} = "TRAIN2"; # MSK 1500Hz tone and single-switch control signal detection.
                             # An empty line is indicated by a fixed frequency signal.
                             # Alternate translation:
                             # Mute if  alternating single tone 
                             # and MSK 1500Hz control signal is detected.
$tone_code{0x24} = "MSK";    # Mute MSK 1200 bps Data (BIIS / MDC1200 ?)

my @skip_code = ("OFF","skip","Pskip");
my %pol_code;
$pol_code{0x00} = "NN";
$pol_code{0x04} = "NR";
$pol_code{0x08} = "RN";
$pol_code{0x0c} = "RR";

my %dsql_type;
$dsql_type{0x00} = "OFF";
$dsql_type{0x01} = "DBEEP";
$dsql_type{0x02} = "DSQL";
$dsql_type{0x03} = "CBEEP";
$dsql_type{0x04} = "CSQL";

$| = 1;

sysopen(CMD,"/dev/tty",O_RDONLY|O_NDELAY) || die "Can't create CMD input filehandle\n";
system("stty -echo -echoe min 1 -icanon -ignpar -inpck -istrip -ixon -ixoff");

if ($help) {
  &do_help;
  &myexit;
}

# Commands:
my %cmd;
my $lead_in = "fe" x400; # 400 'fe' chars.
$cmd{interrogate} = "${lead_in}fefe018019fd";
$cmd{getmemApre} = pack("H*","fefe01801a0001"); #append mem channel
$cmd{getmemBpre} = pack("H*","fefe02801a0001"); #append mem channel

my $xon = 0;  #XON/XOFF flow control state 0=OK 1=XON,don't send anything to the radio.

my $msg_queue = ""; # queue for incoming message
my $msg_call = "";  # call sign of sender of incoming message
my $new_msg = 0;    # 1=Just got a new message, 0=non-new message may be in queue.
my $msg_prefix = ""; # prefix outgoing messages with this string.

#
# main
#

#system("stty $baud -raw -echo min 1 -icanon -isig -ignpar -inpck -istrip -icrnl -ixon -ixoff -opost -iexten <$port");

#open(RADIO,"+<$port") || die "Can't open port $port\n";
if( ! sysopen(RADIO,$port,O_RDWR|O_NDELAY) ) {
 warn "Can't open port $port\n";
 &myexit;
}
system("stty -F $port $baud raw -echo");
$radio_open = 1;


&get_ID;
my $my_call = &get_mycall;
#my $xmsg = &recv("fefe",1);
#my $xmsg = &recv("fefe",1);
#my $xmsg = &recv("fefe",1);
#my $xmsg = &recv("fefe",1);
#my $xmsg = &recv("fefe",1);

&talk();

&end();

sub end {
    print "\nRestoring settings...\n";
    &set_tx("ptt");
    &set_psave("auto");
    close(RADIO);
    &myexit;
}

# final program exit, called when there is a fatal error, or
# by regular exit. Sets tty back to sane mode.
sub myexit {
    close(CMD);
    system("stty cooked echo echoe");
    print "\r\n";
    exit 0;
}

sub talk {
    my $text;
    my $h_text;
    my $m_len;
    my $prompt = "$my_call> ";
    my $c;
    my $buf = "";
    $text = "";
    my $poll_time = time + 2; # at least every 2 seconds
    print $prompt;
    while ( 1 ) {
        $buf = "";
        sysread(CMD,$buf,255);
        for (my $i=0; $i<length($buf); $i++) {
            $c = substr($buf,$i,1);
            if ($c eq "\n") {
                # got a full input line, send it.
                print STDERR "Got input: $buf\n" if ($debug);
                if (length($text) > 0) {
                    # check for command
                    if ( $text =~ /^\// ) {
                        # handle command
                        &do_cmd($text);
                    }
                    else {
                        &send_msg($text);
                    }
                    $text = "";
                }
                $prompt = "$my_call> ";  # reset-prompt (may have changed)
                print "\n$prompt";
            }
            elsif ($c eq "\b" || $c eq "") {
                if (length($text)) { #erase last char.
                    # handle backspace/delete
                    chop $text;
                    &cmd_erase(1);
                }
            }
            elsif ( $c lt " ") {
              # ignore control characters
              $c = "";
            }
            else {
                $text .= $c;
                print $c;
            }
            $c = "";
        }
        my $msg = &recv("fefe80",1);
        #print "R: $msg\n" if(length($msg));
        if(length($msg)) {
            # format message
            $new_msg = 0; # &format_msg sets this to one if we get a new one.
            $msg = &format_msg($msg);
            if (length($msg)) {
                # erase command line
                &cmd_erase(length($prompt) + length($text));

                my $ts = &get_timestamp();
                # if there is a partial message in the message queue
                # it probably needs to get displayed.
                # TODO - perhaps only for certain types of messages?
                if (length($msg_queue) > 0 && $new_msg == 0) {
                    print "$ts $msg_call> $msg_queue\n";
                    $msg_queue = "";
                }
                # display message
                print "$ts $msg\n";
                # re-draw command line
                #print "\nprompt: $prompt\n text: $text\n";
                print "$prompt$text";
            }
        }
        else {
          # may need to poll if nothing coming in from the radio
          if ( time >= $poll_time ) {
              #&rx_poll;
              # poll by getting s-meter setting
              &get_s_meter;

              $poll_time = time + 2;
          }
          &ms_sleep(0.1);
        }
    }
}

sub cmd_erase {
    my $len = shift;
    while ($len > 0) {
        print "\b \b";
        $len--;
    }
}


sub do_cmd {
    my $cmd = shift;

    if ($cmd =~ /exit/i || $cmd =~ /^\/q/i) {
        &end();
    }
    elsif ($cmd =~ /^\/\!/) {
        my $run_cmd = substr($cmd,2);
        print "\nRunning: $run_cmd\n";
        system($run_cmd);
    }
    elsif ($cmd =~ /\/show my/i) {
        &get_mycall("v");
    }
    elsif ($cmd =~ /\/show freq/i) {
        my $freq = &get_freq();
        print "\nFrequency: $freq\n";
    }
    elsif ($cmd =~ /\/show mode/i) {
        my $mode = &get_mode();
        print "\nMode: $mode\n";
    }
    elsif ($cmd =~ /\/show dup/i) {
        my $dup = &get_duplex();
        print "\nDuplex: $dup\n";
    }
    elsif ($cmd =~ /\/show band/i) {
        my $band = &get_band();
        print "\nBand: $band\n";
    }
    elsif ($cmd =~ /\/show tone/i) {
        my $tone = &get_tone();
        print "\nTone: $tone\n";
    }
    elsif ($cmd =~ /\/show rtone/i) {
        my $rtone = &get_rtone(0);
        print "\nRepeater Tone: $rtone\n";
    }
    elsif ($cmd =~ /\/show tsql/i) {
        my $rtone = &get_rtone(1);
        print "\nTSQL Frequency: $rtone\n";
    }
    elsif ($cmd =~ /\/show dtcs/i) {
        my $rtone = &get_rtone(2);
        print "\nDTCS Code: $rtone\n";
    }
    elsif ($cmd =~ /\/show dsql/i) {
        my $dsql = &get_dsql();
        print "\nDSQL: $dsql\n";
    }
    elsif ($cmd =~ /\/show mem/i) {
        my $mem = &get_mem();
    }
    elsif ($cmd =~ /\/show vfo/i) {
        my $freq = &get_vfo();
        print "\nVFO Frequency: $freq\n";
    }
    elsif ($cmd =~ /\/show offset/i) {
        my $freq = &get_offset();
        print "\nOffset Value: $freq Mhz\n";
    }
    elsif ($cmd =~ /\/show power/i) {
        &show_pwr();
    }
    elsif ($cmd =~ /\/show psave/i) {
        &show_psave();
    }
    elsif ($cmd =~ /\/show short/i) { #shortmsg
        my $msg = &get_short_msg();
        print "\nShort Message: \"$msg\"\n";
    }
    elsif ($cmd =~ /\/show gpsmsg/i) {
        my $gpsmsg = &get_gps_msg();
        print "\nGPS Message: $gpsmsg\n";
    }
    elsif ($cmd =~ /\/show gps/i) {
        my $gps = &get_gps();
        print "\nGPS Mode: $gps\n";
    }
    elsif ($cmd =~ /\/show emr/i) {
        my $emr = &get_emr();
        print "\nEMR Mode: $emr\n";
    }
    elsif ($cmd =~ /\/show bk/i) {
        my $bk = &get_bk();
        print "\nBK Mode: $bk\n";
    }
    elsif ($cmd =~ /\/show att/i) {
        my $att = &get_att();
        print "\nAttenuator: $att\n";
    }
    elsif ($cmd =~ /\/show bat/i) {
        my $power_stat = &get_battery();
        print "\nPower Status: $power_stat\n";
    }
    elsif ($cmd =~ /\/show busylo/i) {
        my $bl = &get_busy_lockout();
        print "\nBusy Lockout: $bl\n";
    }
    elsif ($cmd =~ /\/show sql/i) {
        my $sql = &get_sql();
        print "\nSquelch level: $sql\n";
    }
    elsif ($cmd =~ /\/show vol/i) {
        my $vol = &get_vol();
        print "\nVolume level: $vol\n";
    }
    elsif ($cmd =~ /\/show backlight/i)  {
        my $bl = &get_backlight();
        print "\nBacklight: $bl\n";
    }
    elsif ($cmd =~ /\/show rx/i || $cmd =~ /\/heard/) {
        &get_rxcall();
    }
    elsif ($cmd =~ /\/show tx/i) {
        &get_mycall("vtx");
    }
    elsif ($cmd =~ /\/show msgpre/i) {
        print "\nMessage Prefix: \"$msg_prefix\"\n";
    }
    elsif ($cmd =~ /\/set msgpre/i) {
        my($c,$m,$prefix) = split(' ',$cmd,3);
        &set_msg_prefix($prefix);
    }
    elsif ($cmd =~ /\/set my/i) {
        my($c,$m,$i,$call) = split(' ',$cmd,4);
        &set_mycall($i,$call);
        &get_mycall("v");
    }
    elsif ($cmd =~ /\/set power/i) {
        my($c,$m,$lvl) = split(' ',$cmd,3);
        &set_pwr($lvl);
        &show_pwr();
    }
    elsif ($cmd =~ /\/set psave/i) {
        my($c,$m,$state) = split(' ',$cmd,3);
        &set_psave($state);
        &show_psave();
    }
    elsif ($cmd =~ /\/set short/i) { # shortmsg
        # Even though the radio doesn't complain about setting this,
        # It doesn't actully seem to do anything, since when I read
        # the short message after setting it, it is always blank.
        my($c,$m,$msg) = split(' ',$cmd,3);
        &set_short_msg($msg);
        $msg = &get_short_msg();
        print "\nShort Message: \"$msg\"\n";
    }
    elsif ($cmd =~ /\/set gpsmsg/i) {
        my($c,$m,$gpsmsg) = split(' ',$cmd,3);
        &set_gps_msg($gpsmsg);
        $gpsmsg = &get_gps_msg();
        print "\nGPS Message: $gpsmsg\n";
    }
    elsif ($cmd =~ /\/set gps/i) {
        my($c,$m,$gps) = split(' ',$cmd,3);
        &set_gps($gps);
        $gps = &get_gps();
        print "\nGPS mode: $gps\n";
    }
    elsif ($cmd =~ /\/set emr/i) {
        my($c,$m,$emr) = split(' ',$cmd,3);
        &set_emr($emr);
        $emr = &get_emr();
        print "\nEMR mode: $emr\n";
    }
    elsif ($cmd =~ /\/set bk/i) {
        my($c,$m,$bk) = split(' ',$cmd,3);
        &set_bk($bk);
        $bk = &get_bk();
        print "\nBK mode: $bk\n";
    }
    elsif ($cmd =~ /\/set att/i) {
        my($c,$m,$att) = split(' ',$cmd,3);
        &set_att($att);
        $att = &get_att();
        print "\nAttenuator: $att\n";
    }
    elsif ($cmd =~ /\/set busylo/i) {
        my($c,$m,$bl) = split(' ',$cmd,3);
        &set_busy_lockout($bl);
        $bl = &get_busy_lockout();
        print "\nBusy Lockout: $bl\n";
    }
    elsif ($cmd =~ /\/set sql/i) {
        my($c,$m,$sql) = split(' ',$cmd,3);
        &set_sql($sql);
        $sql = &get_sql();
        print "\nSquelch level: $sql\n";
    }
    elsif ($cmd =~ /\/set vol/i) {
        my($c,$m,$vol) = split(' ',$cmd,3);
        &set_vol($vol);
        $vol = &get_vol();
        print "\nVolume level: $vol\n";
    }
    elsif ($cmd =~ /\/set backlight/i) {
        my($c,$m,$bl) = split(' ',$cmd,3);
        &set_backlight($bl);
        $bl = &get_backlight();
        print "\nBacklight: $bl\n";
    }
    elsif ($cmd =~ /\/set tx/) {
        my($c,$m,$tx_call_info) = split(' ',$cmd,3);
        &set_txcall($tx_call_info);
        &get_mycall("vtx"); # show change
    }
    elsif ($cmd =~ /\/set mem/) {
        my($c,$m,$chan) = split(' ',$cmd,3);
        &set_mem($chan);
    }
    elsif ($cmd =~ /\/set vfo/) {
        my($c,$m,$freq) = split(' ',$cmd,3);
        &set_vfo($freq);
        $freq = &get_vfo();
        print "\nVFO Frequency: $freq\n";
    }
    elsif ($cmd =~ /\/set offset/) {
        my($c,$m,$offset) = split(' ',$cmd,3);
        &set_offset($offset);
        $offset = &get_offset();
        print "\nOffset Value: $offset Mhz\n";
    }
    elsif ($cmd =~ /\/set rtone/) {
        my($c,$m,$rtone) = split(' ',$cmd,3);
        &set_rtone($rtone,0);
        $rtone = &get_rtone(0);
        print "\nRepeater Tone: $rtone\n";
    }
    elsif ($cmd =~ /\/set tsql/) {
        my($c,$m,$rtone) = split(' ',$cmd,3);
        &set_rtone($rtone,1);
        $rtone = &get_rtone(1);
        print "\nTSQL Frequency: $rtone\n";
    }
    elsif ($cmd =~ /\/set dtcs/) {
        my($c,$m,$rtone) = split(' ',$cmd,3);
        &set_rtone($rtone,2);
        $rtone = &get_rtone(2);
        print "\nDTCS Code: $rtone\n";
    }
    elsif ($cmd =~ /\/set mode/) {
        my($c,$m,$mode) = split(' ',$cmd,3);
        &set_mode($mode);
        $mode = &get_mode();
        print "\nMode: $mode\n";
    }
    elsif ($cmd =~ /\/set dup/) {
        my($c,$m,$dup) = split(' ',$cmd,3);
        &set_duplex($dup);
        $dup = &get_duplex();
        print "\nDuplex: $dup\n";
    }
    elsif ($cmd =~ /\/set band/) {
        my($c,$m,$band) = split(' ',$cmd,3);
        &set_band($band);
        $band = &get_band();
        print "\nBand: $band\n";
    }
    elsif ($cmd =~ /\/set tone/) {
        my($c,$m,$tone) = split(' ',$cmd,3);
        &set_tone($tone);
        $tone = &get_tone();
        print "\nTone: $tone\n";
    }
    elsif ($cmd =~ /\/set dsql/) {
        my($c,$m,$dsql,$dcode) = split(' ',$cmd,4);
        &set_dsql($dsql,$dcode);
        $dsql = &get_dsql();
        print "\nDSQL: $dsql\n";
    }
    elsif ($cmd =~ /\/say /) {
        my($c,$text) = split(' ',$cmd,2);
        &send_msg($text);
    }
    elsif ($cmd =~ /\/rcmdx/) {  # send raw command to radio with ASCII outptu
        my($c,$msg) = split(' ',$cmd,2);
        &send_debugx_cmd($msg);
    }
    elsif ($cmd =~ /\/rcmd/) {  # send raw command to radio
        my($c,$msg) = split(' ',$cmd,2);
        &send_debug_cmd($msg);
    }
    elsif ($cmd =~ /\/set debug/) {  # Turn on debug output
        my($c,$mode) = split(' ',$cmd,2);
        &set_debug($mode);
    }
    elsif ($cmd =~/\/help/i) { # help
        my($c,$text) = split(' ',$cmd,2);
        &do_help($text);
    }
    else {
        print "\nUnknown command: $cmd\n";
    }
}

sub format_msg {
    my $msg = shift;

    # TODO - Eventually we want to just handle what we want to,
    #        and ignore the rest.


    # use set $new_msg=1 for any cases below that
    # we don't want to print and clear the message queue.

    #fe fe 80 01 1c 00 00 fd
    #PTT readout
    #Ignore
    if ($msg =~ /^fefe800.1c00/) {
        return "";
    }

    #fe fe 80 02 15 01 01 fd
    #Squelch open
    #fe fe 80 02 15 01 00 fd
    #Squelch closed

    #Ignore Squelch status
    if ($msg =~ /^fefe800.1501/) {
        return "";
    }

    #fe fe 80 02 15 02 02 55 fd
    #S-meter level 0255
    #fe fe 80 02 15 02 00 00 fd
    #S-meter level 0

    # ignore S-meter status
    if ($msg =~ /^fefe800.1502/) {
        return "";
    }

    #fe fe 80 02 1d 00 00 00 01 fd
    #D-Star header flag, top=00 bottom=01
    if ($msg =~ /^fefe800.1d0000/) { # ignore
        my $flags = substr($msg,14,4);
        return "D*Flags: $flags";
    }

    #fe fe 80 02 1d 04 00 01 ... fd
    #4b463652414c2042 4b463652414c2042 4b47375749202020 4b463652414c2042 20202020
    #call-sign readout rpt2(8),rpt1(8),your(8),slash(4)
    if ($msg =~ /^fefe800.1d04..../) {
        my $rpt2 = pack("H*",substr($msg,16,16));
        my $rpt1 = pack("H*",substr($msg,32,16));
        my $urcall = pack("H*",substr($msg,48,16));
        my $mycall = pack("H*",substr($msg,64,16));
        my $stroke = pack("H*",substr($msg,80,8));
        $rpt2 =~ s/\s*$//g;
        $rpt1 =~ s/\s*$//g;
        $urcall =~ s/\s*$//g;
        $mycall =~ s/\s*$//g;
        $stroke =~ s/\s*$//g;
        if( length($stroke) ) {
            $mycall .= "/$stroke";
        }
        if (length($msg_queue) > 0 && length($mycall) > 0) {
            # currently receiving a message.
            # save the callsign that goes with it.
            $msg_call = $mycall;
            # don't display this info
            return "";
        }
        my @addr = ($rpt2,$rpt1,$mycall);
        $msg = "";
        for my $part (@addr) {
            if (length($part)) {
                if (length($msg)) {
                    $msg .= ".";
                }
                $msg .= $part;
            }
        }
        $msg =~ s/ /_/g;
        if( length($msg) == 0) {
            $msg = "?";
        }
        if (length($urcall) == 0) {
            $urcall = "?";
        }
        $msg .= " -> $urcall";
        return "ID: $msg";
    }

    #fe fe 80 02 1d c0
    #Received message
    if ( $msg =~ /^fefe800.1dc0/ ) {
        $new_msg = 1; # flag we got a new message.
        my $in_msg_len = substr($msg,12,2);
        $in_msg_len = hex($in_msg_len);
        my $in_msg = pack("H*",substr($msg,14,$in_msg_len * 2));
        $msg_queue .= $in_msg;
        # see if we have a full line that we can output.
        if ($msg_queue =~ /\n/) {
            my $full_msg = "";
            if ($msg_queue =~ /\n$/) {
                $full_msg = $msg_queue;
                $msg_queue = "";
            }
            else {
                # got a full line, and part of another.
                # so return the full line and queue the remainder.
                $msg_queue =~ /^(.*\n)(.*)$/; # want greedy match here.
                $full_msg = $1;
                $msg_queue = $2;
            }
            chomp $full_msg;
            $full_msg =~ s/\r//g; # remove any embedded CRs
            return $full_msg;
        }
        #&get_rx_msg_callsign();
        return ""; # nothing to display yet.
    }

    #Received message readout
    #fefe80021ddf6a6f686e40686179732e6f7267202020202020204b375645202020204a4f484efd
    # (translates to: "john@hays.org       K7VE    JOHN"
    # TODO: Reformat to indicate call sign (starts at pos 21)
    if ( $msg =~ /^fefe800.1ddf/ ) {
        my $short_msg = pack("H*",substr($msg,12,64));
        #&get_rx_msg_callsign();
        return "Short MSG: $short_msg";
    }

    #fe fe 80 02 1d 13 00 fd
    #Auto-display of rx callsign readout
    if ($msg =~ /^fefe800.1d13/) { #ignore
        return "";
    }

    # don't know what this one is
    if ($msg =~ /^fefe800.1d1b/) { #ignore
        return "";
    }

    # Got new TX call setting (auto-changed).
    if ($msg =~/^fefe80021d05/) {
        my $info = &new_txcall($msg);
        return "<TX callsign auto change: $info>";
    }

    #fe fe 80 0? 1d 01 00 00 fd
    #DSQL setting 
    # first byte: 00=Off 01=DBeep 02=DSQL 03=CBeep 04=CSQL
    # 2nd byte: 01=Call Rx 00=No RX
    if ($msg =~ /^fefe80021d01/) { #ignore
        return "";
    }


    #fe fe 80 02 fb fd
    #ACK
    # ignore ACKs
    if ($msg =~ /^fefe800.fbfd/) {
        return "";
    }

    # memory info:
    if ($msg =~ /^fefe800.1a0405/) {
        return &decode_mem($msg);
    }

    # memory bank:
    if ($msg =~ /^fefe800.1a0409/) {
        #my $bank = pack("H*",substr($msg,14,2));
        #return "Memory Bank: $bank";
        return "";  # ignore this, it seems to be redundant.
    }

    # band A/B
    if ($msg =~ /^fefe80011a0401/) {
      my $band = &decode_band($msg);
      return "Band $band";
    }

    # single/dual band
    if ($msg =~ /^fefe80011a0400/) {
      return &decode_band_mode($msg);
    }

    # frequency:
    if ($msg =~ /^fefe800.00/) {
        my $b5 = substr($msg,10,2);
        my $b4 = substr($msg,12,2);
        my $b3 = substr($msg,14,2);
        my $b2 = substr($msg,16,2);
        my $b1 = substr($msg,18,2);
        my $freq = "$b1$b2$b3$b4$b5";
        $freq /= 1000000;
        return "Frequency: $freq Mhz";
    }

    # ignore volume changes
    if ($msg =~ /^fefe800.1401/) {
        return "";
    }

    # Power turned off!
    if ($msg =~ /^fefe80011800fd/) {
      return "Radio turned off!";
    }

    # Power status change
    if ($msg =~ /^fefe80011a040e/) {
      my $pstat = &decode_battery($msg);
      return "Power Status: $pstat";
    }

    return $msg;
}

sub set_debug {
  my $mode = shift;
  if ( $mode =~ /on/i ) {
    $debug = 1;
  }
  else {
    $debug = 0;
  }
}

#
# set TX to AUTO or PTT
sub set_tx {
    my $mode = shift;
    my $xmode = 0;
    if ($mode =~ /^auto$/i) {
        $xmode = 1; # AUTO
    }
    else {
        $xmode = 0; # PTT
    }
    $mode = sprintf("%02x",$xmode);
    &send2("fefe02801d18${mode}fd");
    #&send("fefe02801d18${mode}fd");
}

#
# send message
sub send_msg {
    my $text = shift;
    $text .= "\r\n";  # append newline, perhaps use \r\n
    if(length($msg_prefix) > 0) {
        $text = "$msg_prefix $text\n";
    }

    #&set_tx("auto"); # now done at start of program.

    my $m_len = length($text);
    #if ($m_len > 255) {
    #    warn "\nLong message truncated\n";
    #    $m_len = 255;
    #}
    my $send_text = $text;
    my $psize = 32; # max=32 #dChat seems to use 5.
    while ( length($send_text) > $psize ) {
        my $sub_text = substr($send_text,0,$psize);
        my $m_hex = unpack("H*",substr($sub_text,0,$psize));
        my $m_lx = sprintf("%02x",$psize);
        #print "Send msg piece\n";
        &send2("fefe02801dc1${m_lx}${m_hex}fd");
        $send_text = substr($send_text,$psize);
    }
    $m_len = length($send_text);
    my $m_hex = unpack("H*",substr($send_text,0,$m_len));
    my $m_lx = sprintf("%02x",$m_len);
    #&send2("fefe02801dc1${m_lx}${m_hex}fd");
    #&send("fefe02801dc1${m_lx}${m_hex}fd");
    #print "\nSend:fefe02801dc1${m_lx}${m_hex}fd\n";
    &send2("fefe02801dc1${m_lx}${m_hex}fd");

    # Note - Some delay is needed before AUTO PTT is turned off.
    #        I tried turning it on and off for each message, but
    #        that resulted in it getting turned off and TX stopping
    #        before the entire message was transmitted.
    # TX is now set to 'auto' at program start, and set back
    # to PTT at exit now.  Seems to solve this problem.
    # Old delay code is below for reference.
    #print "\nLast packet size: $m_len\n";
    #&ms_sleep(0.45); # 300 milliseconds for 32 characters?
    # 5ms for 5 characters?
    # sleep time = $packet_size / 100?
    #&set_tx("ptt");
}

#
# check for received messages
sub rx_poll {
    print "RX Poll\n" if ($debug);
    while (1) {
        # print STDERR "\nrx_poll loop\n";
        my $msg = &recv("fefe80",1);
        # print STDERR "\nrx_poll loop after recv(fefe80)\n";
        if (length($msg) == 0) {
            # print STDERR "get sql\n";
            &get_s_meter;
            # print STDERR "\nexit rx_poll\n";
            return;
        }

        # got something. Get full message
        # print STDERR "Read full message\n";
        if ( $msg !~ /fd$/ ) {
            my $msg2 = &read_data;
            $msg .= $msg2;
        }
        print "\nReceived: $msg\n" if ($debug);
        if ($msg =~ /^021dc0/ ) {
            # Received a data message
        }
    }
}

sub get_ID {
  # make sure we are talking, and talking to an IC-91A/D
  &send($cmd{interrogate});
  my $msg = &recv("fefe800119",0);
  #expecting: fefe80011928880100050f02017780fd
  my $model = substr($msg,10,6);
  if ($model eq "288801") {  #IC-92AD might be "306601"
    print STDERR "Connected\n";
  }
  else {
    warn "Invalid radio model: $model\n";
    &myexit;
  }
  &power_on;
  &set_tx("auto");
  $psave_state = &get_psave;
  &set_psave("off");
}

sub power_on {
    # send power on command
    &send2("fefe01801801fd");
    #print "Power on cmd OK\n";
}

sub get_s_meter {
    &send("fefe01801502fd");
    my $msg = &recv("fefe80011502",0);
}


sub read_data {
    my $msg = "";
    my $msg_end = 0;
    while ( ! $msg_end ) {
        my $data = "";
        sysread(RADIO,$data,1);
        if (length($data) == 0) {
            #sleep(1); # Don't sleep this long. Maybe 100ms?
            &ms_sleep(0.1);
            next;
        }
        my $hx = unpack("H*",$data);
        $msg .= $hx;
        print STDERR "$hx" if ($debug);
        last if ($hx eq "fd");
    }
}

#
# Send command to radio and wait for it to be acknowledged.
#
sub send2 {
  my $msg = shift;
  &send($msg);

  # wait for acknowledge
  my $stat = 1;
  while ( 1 ) {
    my $ack = &recv("fefe800",0);
    if ($ack =~ /^fefe800.fbfd/) {
      # OK
      print "ACK OK\n" if ($debug);
      $stat = 0;
      last;
    }
    elsif ($ack =~ /^fefe800.fafd/) {
      # No good
      if ($debug) {
        warn "msg: $msg\n";
      }
      print "ACK NAK\n" if ($debug);
      last;
    }
    elsif ($ack eq "") {
      # Assume NACK
      print "No ACK or NAK received, assuming NAK\n" if ($debug);
      $stat = 1;
      last;
    }
    else {
      # TODO: The unsolicited message received should probably
      #       be processed rather than ignored.
      if ($debug) {
        warn "Unexpected ACK response: $ack (still waiting)\n";
      }
    }
  }
  return $stat;
}

sub send {
  my $msg = shift;
  if ($xon) { # wait for XOFF
      &flow_wait();
  }
  print STDERR "s:$msg\n" if ($debug);
  $msg = pack("H*",$msg);
  my $rv = syswrite(RADIO,$msg);
  warn "Write error\n" if ($rv < 0);
}

sub recv {
  my $expected = shift;
  my $poll = shift;
  my $msg = "";
  while( 1 ) {
    $msg = "";
    my $c = "";
    print STDERR "r:" if ($debug);
    my $first = 1;
    my $poll_cnt = 0;
    while ( 1 ) {
      my $rv = sysread(RADIO,$c,1);
      if ($rv == 0 || length($c) == 0) {
        return "" if ($poll); # Polling, nothing ready so return
        &ms_sleep(0.1); # wait for input
        $poll_cnt++;
        if ($poll_cnt > 80) {
          # went more than 8 seconds without a response,
          # reset communications link and return
          warn "\nRadio not responding...\n";
          &send($cmd{interrogate});
          my $msg = &recv("fefe800119",0);
          if (length($msg) > 0) {
            &power_on;
            print STDERR "Reconnected          ";
            &set_tx("auto");
            &send("fefe01801a0401fd"); # get current band
          }
          return "";
        }
        next;
      }
      my $hx = unpack("H*",$c);

      # skip input characters until we get to the start
      # of a message.
      $xon = 1 if ($first && $hx eq "13");
      $xon = 0 if ($first && $hx eq "11");
      if ($first && $hx ne "fe") {
          next;
      }
      else {
          $first = 0;
      }
      $msg .= $hx;
      print STDERR "$hx" if ($debug);
      last if ($hx eq "fd");
    }
    print STDERR "\n" if ($debug);
    # may get extra characters that need to be chopped.
    if ($msg !~ /$expected/) {
      if ($msg =~/^fefe80..fafd/ ) {
        warn "Data/command rejected by radio\n";
        last;
      }
      #print "Unsolicited response: $msg\n";
      # print "Expected: $expected\n";
    }
    else {
      #print "Found proper response: $msg\n";
      last;
    }
  }
  return $msg;
}

sub flow_wait {
    my $c = "";
    print "Flow wait\n" if ($debug);
    my $cnt = 0;
    while ( 1 ) {
        my $rv = sysread(RADIO,$c,1);
        if($rv == 0 || length($c) == 0) {
            &ms_sleep(0.1);
            $cnt++;
            if ($cnt > 20) {
                print "\nXON timeout\n";
                $xon = 0;
                return;
            }
            next;
        }

        my $hx = unpack("H*",$c);
        if ($hx eq "11") {
            $xon = 0;
            last;
        }
    }
}

sub ms_sleep {
    my $sleep_time = shift;
    #print STDERR "  Sleep $sleep_time \n";
    select(undef,undef,undef,$sleep_time);
}

sub get_timestamp {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    return sprintf("%02d:%02d:%02d",$hour,$min,$sec);
}

sub get_mycall {
    my $opt = shift || "";
    my $tx_call_ref = &get_txcall();
    my $rpt2 = $$tx_call_ref[0];
    my $rpt1 = $$tx_call_ref[1];
    my $urcall = $$tx_call_ref[2];

    &send("fefe02801d02fd"); # Get which entry in 'mycall' table is current.
    my $msg = &recv("fefe80021d02",0);
    my $mycall_num = substr($msg,12,2);
    $mycall_num--;

    &send("fefe02801d08fd"); # Get 'mycall' table
    my @mycalls;
    for (my $i=0; $i<=5; $i++) {
        $mycalls[$i] = &recv("fefe80021d08",0);
    }
    #print "mycalls: $#mycalls\n";
    #print "Sel: $mycalls[$mycall_num]\n";
    my $mycall = &format_mycall($mycalls[$mycall_num]);

    if( $opt eq "v") {
        # show mycall information
        my $sel_call = "";
        print "\n";
        for ( my $i=0;$i<=$#mycalls;$i++) {
            my $t_mycall = &format_mycall($mycalls[$i]);
            my $n = $i + 1;
            $sel_call = "   ";
            $sel_call = "==>" if ($i == $mycall_num);
            print "$sel_call $n: $t_mycall\n";
        }
        print "   TX RPT2:\"$rpt2\" RPT1:\"$rpt1\" URCALL:\"$urcall\"\n";
    }
    elsif( $opt eq "vtx") {
        #print "\nMYCALL:\"$mycall\" RPT2:\"$rpt2\" RPT1:\"$rpt1\" URCALL:\"$urcall\"\n";
        $rpt2 =~ s/ /_/g;
        $rpt1 =~ s/ /_/g;
        $urcall =~ s/ /_/g;
        print "\n$mycall -> $rpt2.$rpt1.$urcall\n";
    }


    #print "TX rpt2: $rpt2\n";
    #print "TX rpt1: $rpt1\n";
    #print "TX urcall $urcall\n";
    #print "TX mycall $mycall\n";
    return $mycall;
}

sub format_mycall {
    my $mycall_data = shift;
    my $mycall = pack("H*",substr($mycall_data,14,24));
    my $myslash = substr($mycall,8,4);
    $mycall = substr($mycall,0,8);
    $mycall =~ s/\s*$//;
    $myslash =~ s/\s*$//;
    if (length($myslash)) {
        $mycall .= "/$myslash";
    }
    return $mycall;
}


sub get_rxcall {
    print "\n   RPT2     RPT1     Dest     Caller        Stat\n";
    print   "-- -------- -------- -------- ------------- ----\n";
    &send("fefe02801d04fd");
    my $n = 0;

    while (1) {
        my $msg = &recv("fefe80021d04",0);
        #print "RXcall: $msg\n";
        my $i = hex(substr($msg,12,2));

        # This byte I believe to contain either BK and EMR indicators or status
        # but I don't know how it is encoded.
        # RS91 shows nothing even though columns contain 01 or 02
        # for some entries.
        my $flg = hex(substr($msg,14,2));
        my $xflg = $flg;
        if( $flg == 2) {
            $xflg = "UR?";
        }
        elsif( $flg == 1) {
            $xflg = "RPT?";
        }

        my $data = pack("H*",substr($msg,16));
        my $rpt2 = substr($data,0,8);
        $rpt2 =~ s/ /_/g;
        my $rpt1 = substr($data,8,8);
        $rpt1 =~ s/ /_/g;
        my $urcall = substr($data,16,8);
        $urcall =~ s/ *$//;
        $urcall =~ s/ /_/g;
        my $mycall = substr($data,24,8);
        $mycall =~ s/ *$//;
        $mycall =~ s/ /_/g;
        my $myslash = substr($data,32,4);
        $n++;
        last if ($n > 20);
        next if ($data =~ /^ {32}/); # ignore empty entries
        printf("%02d %-8.8s %-8.8s %-8.8s %-8.8s/%-4.4s %s\n",
               $i,$rpt2,$rpt1,$urcall,$mycall,$myslash,$xflg);
    }
}

sub set_mycall {
    my $idx = shift;
    my $mycall = shift || "";
    $mycall =~ s/\s*//g;

    if($idx < 1 || $idx > 6) {
        print "\nInvalid mycall number: $idx\n";
        return;
    }

    $idx = sprintf("%02d",$idx);
    if( length($mycall) == 0) {
        # set using index only
        &send2("fefe02801d02${idx}fd");
    }
    else {
        my($call,$slash) = split('/',$mycall,2);
        if($mycall eq "-") { # clear this memory
            $call = "";
            $slash = "";
        }
        $call =~ y/[a-z]/[A-Z]/; #make sure uppercase
        $slash =~ y/[a-z]/[A-Z]/; #make sure uppercase
        $call = sprintf("%-8.8s",$call); # Pad to 8 characters
        $slash = sprintf("%-4.4s",$slash); #Pad to 4 characters
        my $xcall = unpack("H*","$call$slash");
        #print "\nSet mycall:fefe028001d08${idx}${xcall}fd\n";
        &send2("fefe02801d08${idx}${xcall}fd");
    }
    $my_call = &get_mycall();
}

sub get_rx_msg_callsign {
    &send("fefe02801de2fd");
    my $msg = &recv("fefe80021de2",0);
    print "\nRX MSG CS: $msg\n";
    # This just seems to always return '01'. It must be
    # This is actually Tx Message Off/On
    return $msg;
}

sub set_txcall {
    my $xcall = shift;
    my @acall = split('\.',$xcall,3);
    my $urcall = pop @acall;
    my $rpt1 = pop @acall || "";
    my $rpt2 = pop @acall || "";


    $urcall =~ y/a-z/A-Z/;
    $rpt1 =~ y/a-z/A-Z/;
    $rpt2 =~ y/a-z/A-Z/;

    # get current settings.
    my $cur_tx_call_ref = &get_txcall();
    $rpt2 = $$cur_tx_call_ref[0] if ($rpt2 eq "");
    $rpt1 = $$cur_tx_call_ref[1] if ($rpt1 eq "");
    $urcall = $$cur_tx_call_ref[2] if ($urcall eq "");
    $urcall = "CQCQCQ  " if($urcall =~ /^CQ/);

    # change all '_' to spaces
    $urcall =~ s/_/ /g;
    $rpt1 =~ s/_/ /g;
    $rpt2 =~ s/_/ /g;
    #
    # '-' or leading space clears rpt2/rpt1/urcall
    $urcall = "" if ($urcall =~ /^[ -]/);
    $rpt1 = "" if ($rpt1 =~ /^[ -]/);
    $rpt2 = "" if ($rpt2 =~ /^[ -]/);

    my $txcall= sprintf("%-8.8s%-8.8s%-8.8s",$rpt2,$rpt1,$urcall);
    $txcall = unpack("H*",$txcall);
    #print "\nsend: $txcall\n";
    &send2("fefe02801d05${txcall}fd");
    #print "OK\n";
}

sub get_txcall {
    &send("fefe02801d05fd");
    my $msg = &recv("fefe80021d05",0);
    #print "mycall msg: $msg\n";
    return &format_txcall($msg);
}

sub format_txcall {
    my $msg = shift;
    my $rpt2 = pack("H*",substr($msg,12,16));
    my $rpt1 = pack("H*",substr($msg,28,16));
    my $urcall = pack("H*",substr($msg,44,16));
    $rpt2 =~ s/\s*$//g;
    $rpt1 =~ s/\s*$//g;
    $urcall =~ s/\s*$//g;
    my @calldata = ($rpt2,$rpt1,$urcall);
    return \@calldata;
}

sub new_txcall {
    my $msg = shift;
    my $tx_call_ref = &format_txcall($msg);
    my $txcall = "$my_call -> " .
        $$tx_call_ref[0] . "." .
        $$tx_call_ref[1] . "." .
        $$tx_call_ref[2];
    return $txcall;
}

sub set_pwr {
    my $lvl = shift;
    my $band = &get_bandno;
    $lvl = 255 if( $lvl =~ /H/i);
    $lvl = 0 if( $lvl =~ /L/i);
    $lvl = 255 if($lvl > 255);
    $lvl = 0 if ($lvl < 0);
    $lvl = sprintf("%04d",$lvl);
    &send2("fefe${band}80140a${lvl}fd");
}

sub show_pwr {
    my $band = &get_bandno;
    &send("fefe${band}80140afd");
    my $msg = &recv("fefe80${band}140a",0);
    my $lvl = substr($msg,12,4);
    $lvl += 0;
    print "\nTransmitter Power: $lvl\n";
}

sub set_psave {
    my $state = shift;
    my $nstate = $state; 
    $nstate = "00" if ($state =~ /off/i);
    $nstate = "01" if ($state =~ /auto/i);
    &send2("fefe01801a0541${nstate}fd");
}

sub get_psave {
    &send("fefe01801a0541fd");
    my $msg = &recv("fefe80011a0541",0);
    my $nstate = substr($msg,14,2);
    my $state = $nstate;
    $state = "off" if ($nstate eq "00");
    $state = "auto" if ($nstate eq "01");
    return $state;
}

sub show_psave {
    my $state = &get_psave;
    print "\nPower Save: $state\n";
}

sub get_freq {
    my $band = &get_bandno;
    &send("fefe${band}8003fd");
    my $msg = &recv("fefe80${band}03",0);
    my $freq = substr($msg,18,2) . substr($msg,16,2) . "." .
               substr($msg,14,2) . substr($msg,12,2) . " Mhz";
    $freq =~ s/^0*//;
    return $freq;
}

#
# set radio memory channel
#
sub set_mem {
    my $chan = shift || -1;
    # set mem channel:  fefe02801a04040nnnfd
    # set bank channel: fefe02801a04054201fd ?? set to 'B01'
    # set bank: fefe80021a0409xxfd ?? sets bank xx (42=B)
    #
    # Useful commands:
    # set bank mode:   fefe02801a040202fd
    # set memory mode: fefe02801a040201fd
    # set vfo mode:    fefe02801a040200fd
    my $band = &get_bandno;

    if ( $chan =~ /^[A-Z,a-z]/ ) {
        # set memory bank channel
        my $bank = substr($chan,0,1);
        $bank =~ y/a-z/A-Z/;
        my $chan_num = substr($chan,1);
        $chan_num = -1 if(length($chan_num) == 0);
        my $bank = unpack("H*",$bank);
        my $chan_num = sprintf("%02d",$chan_num);
        &send2("fefe${band}801a040202fd"); # set bank mode
        if( $chan_num >= 0) {
            &send2("fefe${band}801a0405${bank}${chan_num}fd");
        }
    }
    else {
        $chan = sprintf("%04d",$chan);
        &send2("fefe${band}801a040201fd"); # set memory mode
        if ($chan >= 0) {
            &send2("fefe${band}801a0404${chan}fd");
        }
    }
}

sub get_mem {
    my $band = &get_bandno;

    # get memory/bank/vfo mode
    &send("fefe${band}801a0402fd");
    my $mode = &recv("fefe80${band}1a0402",0);
    my $mode = substr($mode,14,2);
    if ($mode eq "01") { # memory mode
        &send("fefe${band}801a0404fd");
        my $msg = &recv("fefe80${band}1a0404",0);
        my $mem_no = substr($msg,14,4) + 0;
        &mem_detail($band,$mem_no);
        #print "\nMemory #$mem_no\n";
    }
    elsif ($mode eq "02") { # bank memory mode
        # get current bank: fefe028001a0409fd
        &send("fefe${band}801a0409fd");
        my $bank = &recv("fefe80${band}1a0409",0);
        my $bank = substr($bank,14,2);
        &send("fefe${band}801a0405${bank}fd");
        my $msg = &recv("fefe80${band}1a0405",0);
        my $mem_no = substr($msg,18,4) + 0;

        # Get all memory info (name,freq,mode,tsql...)
        &mem_detail($band,$mem_no);

        #my $info = &decode_mem($msg);
        #print "\n$info\n";
    }
    else {
        print "\nRadio is in VFO mode\n";
    }
}

sub decode_mem {
    my $msg = shift;
    my $bank = pack("H*",substr($msg,14,2));
    my $chan = substr($msg,16,2);
    my $mem_no = substr($msg,18,4) + 0;
    return "Memory: $bank$chan #$mem_no";
}

# show detailed information for a specified memory number.
sub mem_detail {
  my $band = shift;
  my $memno = shift;
  
  $memno = sprintf("%04d",$memno);
  &send("fefe${band}801a0001${memno}fd");
  my $msg = &recv("fefe80${band}1a0001",0);
  my $aref = &decode_full_memory($band,1,$msg);
  my $xmem = sprintf("%03d",$$aref[1]);
  my $xchan = sprintf("%02d",$$aref[15]);
  print "\nBand $$aref[0] Memory #$xmem  Bank $$aref[14]$xchan\n";
  print "     Freq: $$aref[2] \"$$aref[7]\" Mode: $$aref[6]\n";
  print "     $$aref[3] Tone:$$aref[9] RTone:$$aref[10] CTCSS:$$aref[11] DTCS:$$aref[12]\n";
  if ($band eq "02") {
  print "     DSQL:$$aref[16]($$aref[17]) URCALL:\"$$aref[18]\" RPT1:\"$$aref[19]\" RPT2:\"$$aref[20]\"\n";
  }
}

sub decode_full_memory {
  my $band = shift; # 1=A 2=B
  my $mtype = shift; # 1=Memory/PS 2=Call
  my $msg = shift;

  my @memdata;  # return reference to this array

  my $vband = substr($msg,6,2);
  $vband = $vband eq "01" ? "A" : "B";

  my $mem_type = substr($msg,12,2);

  my $rmemch = substr($msg,14,4);
  $rmemch += 0;
  my $last_mem = 800;
  $last_mem = 400 if( $vband eq "B");

  # handle special names of Programmed scan memory
  if ($mem_type eq "01") {
    if ($rmemch >= $last_mem && $rmemch < ($last_mem + 50)) {
      my $psnum = 0;
      my $psmem = $rmemch - $last_mem;
      $psnum = $psmem / 2;
      my $ab = $psmem % 2 == 1 ? "b" : "a";
      $rmemch = sprintf("PS-%02d-%s",$psnum,$ab);
    }
  }
  # Call channel memory - Format memory name
  elsif ($mem_type eq "02") {
    $rmemch = sprintf("Call-%d",$rmemch);
  }
  elsif ($mem_type eq "00") {
    $rmemch = sprintf("V-%02d",$rmemch);
  }

  if ( substr($msg,18,2) eq "ff" ) {
    # nothing in here, it's empty
    return \@memdata;
  }
  my $freq = substr($msg,26,2) . substr($msg,24,2) . substr($msg,22,2) .
             substr($msg,20,2) . substr($msg,18,2);
  $freq /= 1000000;
  $freq = sprintf("%8.6f",$freq);

  my $offset = substr($msg,36,2) . substr($msg,34,2) . substr($msg,32,2) . 
               substr($msg,30,2) . substr($msg,28,2);
  $offset /= 1000000;
  $offset = sprintf("%8.6f",$offset);

  my $tone = substr($msg,38,4);
  $tone /= 10;
  $tone = sprintf("%5.1f",$tone);
  my $ctcss = substr($msg,42,4);
  $ctcss /= 10;
  $ctcss = sprintf("%5.1f",$ctcss);

  my $dtcs = substr($msg,46,4);
  $dtcs += 0;

  my $train_freq = substr($msg,50,4);

  my $mode = substr($msg,54,1);
  $mode = $mod_mode[$mode];

  my $tstep = substr($msg,55,1);
  $tstep = hex($tstep);
  $tstep = $tstep_code[$tstep];

  my $dup = hex(substr($msg,57,1)) & 0x03;
  $dup = $dup_code[$dup];

  my $tone_type = hex(substr($msg,56,2));
  $tone_type &= 0x1c;
  $tone_type = $tone_code{$tone_type};

  my $skip = hex(substr($msg,58,2)) & 0x03;
  $skip = $skip_code[$skip];

  my $polarity = hex(substr($msg,58,2)) & 0x0c;
  $polarity = $pol_code{$polarity};

  my $bank = substr($msg,60,2);
  if ($bank eq "00") {
    # No bank assigment for this memory channel
    $bank = "";
  }
  else {
    $bank = pack("H*",$bank);
  }
  my $bchan = substr($msg,62,2);
  $bchan += 0;

  my $chan_name = pack("H*",substr($msg,64,16));

  my $urcall = "";
  my $rpt1 = "";
  my $rpt2 = "";
  my $dsql = "";
  my $dsql_code = "";
  if( $band == 2 ) {
    $dsql = hex(substr($msg,80,2)) & 0x07;
    $dsql = $dsql_type{$dsql};

    $dsql_code = substr($msg,82,2);

    $rpt2 = pack("H*",substr($msg,84,16));
    $rpt1 = pack("H*",substr($msg,100,16));
    $urcall = pack("H*",substr($msg,116,16));
  }

  @memdata = (
        $vband,
        $rmemch,
        $freq,
        $dup,
        $offset,
        $tstep,
        $mode,
        $chan_name,
        $skip,
        $tone_type,
        $tone,
        $ctcss,
        $dtcs,
        $polarity,
        $bank,
        $bchan,
        $dsql,
        $dsql_code,
        $urcall,
        $rpt1,
        $rpt2,
        $train_freq
  );
  return \@memdata;
}
sub set_vfo {
    my $freq = shift;
    my $band = &get_bandno;
    if ( $freq =~ /^mem/i ) {
      # copy current memory to VFO
      &send2("fefe${band}800afd"); # copy mem to VFO
      &send2("fefe${band}801a040200fd"); # set VFO mode
    }
    else {
      $freq *= 1000000;  # convert to HZ
      $freq = sprintf("%010d",$freq);
      my $b1 = substr($freq,0,2);
      my $b2 = substr($freq,2,2);
      my $b3 = substr($freq,4,2);
      my $b4 = substr($freq,6,2);
      my $b5 = substr($freq,8,2);

      # put radio into VFO mode
      &send2("fefe${band}801a040200fd");

      if ($freq > 0) {
      # set VFO frequency
          &send2("fefe${band}8005${b5}${b4}${b3}${b2}${b1}fd");
      }
    }
}

sub get_vfo {
    my $band = &get_bandno;
    &send("fefe${band}8003fd");
    my $msg = &recv("fefe80${band}03",0);
    my $b5 = substr($msg,10,2);
    my $b4 = substr($msg,12,2);
    my $b3 = substr($msg,14,2);
    my $b2 = substr($msg,16,2);
    my $b1 = substr($msg,18,2);
    my $freq = "$b1$b2$b3$b4$b5";
    $freq /= 1000000;
    return $freq;
}

sub set_mode {
    my $mode = shift;
    my $band = &get_bandno;
    my $cmd = "";
    if ($mode =~ /^FM$/i) {
        $cmd = "0501";
    }
    elsif ($mode =~ /^FM-N$/i) {
        $cmd = "0502";
    }
    elsif ($mode =~ /^WFM$/i) {
        $cmd = "0601";
    }
    elsif ($mode =~ /^AM$/i) {
        $cmd = "0201";
    }
    elsif ($mode =~ /^DV$/i) {
        $cmd = "d001";
    }
    elsif ($mode =~ /^DD$/i) {
        $cmd = "d101";
    }
    else {
        print "\nInvalid mode: $mode";
        return;
    }
    &send2("fefe${band}8006${cmd}fd");
}

sub get_mode {
    my $band = &get_bandno;
    &send("fefe${band}8004fd");
    my $msg = &recv("fefe80${band}04",0);
    my $mode = substr($msg,10,4);
    my $mmode = "";
    $mmode = "FM" if ($mode eq "0501");
    $mmode = "FM-N" if ($mode eq "0502");
    $mmode = "WFM" if ($mode eq "0601");
    $mmode = "AM" if ($mode eq "0201");
    $mmode = "DV" if ($mode eq "d001");
    $mmode = "DD" if ($mode eq "d101");
    return $mmode;
}

#
# send a user specified command to the radio (for debugging only)
#
sub send_debug_cmd {
  my $cmd = shift;
  &send("fefe${cmd}fd");
  my $msg = &recv("fefe80",0);
  while ( $msg =~ /^fefe80011502/ ) {
    # ignore S-meter status and try again
    $msg = &recv("fefe80",0);
  }
  print "\n Initial Response: $msg\n";
}
#
# send a user specified command to the radio (for debugging only)
# Translate the response to ASCII
#
sub send_debugx_cmd {
  my $cmd = shift;
  &send("fefe${cmd}fd");
  my $msg = &recv("fefe80",0);
  while ( $msg =~ /^fefe80011502/ ) {
    # ignore S-meter status and try again
    $msg = &recv("fefe80",0);
  }
  print "\n Initial Response: $msg\n";
  my $resp = $cmd;
  $resp =~ s/^(\d\d)(\d\d)/$2$1/;
  #print "Command:  $cmd\n";
  #print "Response: $resp\n";
  if ( $msg =~ /$resp/ ) {
    my $submsg = $msg;
    $submsg =~ s/fd$//;
    $submsg = substr($submsg,length($resp) + 4);
    $submsg = pack("H*",$submsg);
    print "ASCII: $submsg\n";
  }
}


sub set_duplex {
    my $dup = shift;
    my $xdup = "10";
    my $band = &get_bandno;
    $xdup = "11" if ($dup =~ /\-/);
    $xdup = "12" if ($dup =~ /\+/);
    &send2("fefe${band}800f${xdup}fd");
}

sub get_duplex {
    my $band = &get_bandno;
    &send("fefe${band}800ffd");
    my $msg = &recv("fefe80${band}0f",0);
    my $xdup = substr($msg,10,2);
    my $dup = "Unknown $xdup";
    $dup = "-" if ($xdup eq "11");
    $dup = "+" if ($xdup eq "12");
    $dup = "Simplex" if ($xdup eq "10");
    return $dup;
} 

sub set_tone {
    my $tone = shift;
    my $xtone = "00";
    my $band = &get_bandno;
    $xtone = "00" if ( $tone =~ /off/i );
    $xtone = "01" if ( $tone =~ /tone/i );
    $xtone = "02" if ( $tone =~ /tsqlb/i ); # not setable?
    $xtone = "03" if ( $tone =~ /tsql/i );
    $xtone = "04" if ( $tone =~ /dtcsb/i ); # not setable?
    $xtone = "05" if ( $tone =~ /dtcs/i );
    $xtone = "06" if ( $tone =~ /tsqlr/i );
    $xtone = "07" if ( $tone =~ /train1/i );
    $xtone = "08" if ( $tone =~ /train2/i );
    $xtone = "09" if ( $tone =~ /MSK/i );
    &send2("fefe${band}801a02${xtone}fd");
}

sub get_tone {
    my $band = &get_bandno;
    &send("fefe${band}801a02fd");
    my $msg = &recv("fefe80${band}1a02",0);
    my $xtone = substr($msg,12,2);
    my $tone = "Unknown $xtone";
    $tone = "OFF" if ($xtone eq "00");
    $tone = "TONE" if ($xtone eq "01");
    $tone = "TSQL PBEEP" if ($xtone eq "02");
    $tone = "TSQL" if ($xtone eq "03");
    $tone = "DTCS PBEEP" if ($xtone eq "04");
    $tone = "DTCS" if ($xtone eq "05");
    $tone = "TSQL R" if ($xtone eq "06");
    $tone = "TRAIN1" if ($xtone eq "07");
    $tone = "TRAIN2" if ($xtone eq "08");
    $tone = "MSK" if ($xtone eq "09");
    return $tone;
}

sub set_dsql {
    my $tone = shift;
    my $dcode = shift || "";
    if (length($dcode) > 0) {
        &set_dcode($dcode);
    }
    my $xtone = "00";
    $xtone = "00" if ( $tone =~ /off/i );
    $xtone = "01" if ( $tone =~ /dsqlb/i );
    $xtone = "02" if ( $tone =~ /dsql/i );
    $xtone = "03" if ( $tone =~ /csqlb/i );
    $xtone = "04" if ( $tone =~ /csql/i );
    &send2("fefe02801d01${xtone}fd");
}

sub get_dsql {
    &send("fefe02801d01fd");
    my $msg = &recv("fefe80021d01",0);
    my $xtone = substr($msg,12,2);
    my $tone = "Unknown $xtone";
    my $dcode = &get_dcode();
    $tone = "OFF" if ($xtone eq "00");
    $tone = "DSQL PBEEP $dcode" if ($xtone eq "01");
    $tone = "DSQL $dcode" if ($xtone eq "02");
    $tone = "CSQL PBEEP" if ($xtone eq "03");
    $tone = "CSQL" if ($xtone eq "04");
    return $tone;
}

sub set_dcode {
    my $dcode = shift;
    $dcode = sprintf("%02d",$dcode);
    &send2("fefe02801d17${dcode}fd");
}

sub get_dcode {
    &send("fefe02801d17fd");
    my $msg = &recv("fefe80021d17",0);
    my $dcode = substr($msg,12,2);
    return $dcode;
}

sub set_offset {
    my $offset = shift;
    my $band = &get_bandno;
    $offset *= 1000000;  # convert to HZ
    $offset = sprintf("%010d",$offset);
    my $b1 = substr($offset,0,2);
    my $b2 = substr($offset,2,2);
    my $b3 = substr($offset,4,2);
    my $b4 = substr($offset,6,2);
    my $b5 = substr($offset,8,2);

    # set frequency offset value
    &send2("fefe${band}800d${b5}${b4}${b3}${b2}${b1}fd");
}

sub get_offset {
    my $band = &get_bandno;
    &send("fefe${band}800cfd");
    my $msg = &recv("fefe80${band}0c",0);
    my $b5 = substr($msg,10,2);
    my $b4 = substr($msg,12,2);
    my $b3 = substr($msg,14,2);
    my $b2 = substr($msg,16,2);
    my $b1 = substr($msg,18,2);
    my $offset = "$b1$b2$b3$b4$b5";
    $offset /= 1000000;
    return $offset;
}

sub set_rtone {
    my $rtone = shift;
    my $tone_type = shift;
    my $band = &get_bandno;
    $tone_type = sprintf("%02d",$tone_type);
    $rtone *= 10 if ($tone_type < 2);
    $rtone = sprintf("%04d",$rtone);
    &send2("fefe${band}801b${tone_type}${rtone}fd");
}

sub get_rtone {
    my $tone_type = shift;
    my $band = &get_bandno;
    $tone_type = sprintf("%02d",$tone_type);
    &send("fefe${band}801b${tone_type}fd");
    my $msg = &recv("fefe80${band}1b${tone_type}",0);
    my $tone_freq = substr($msg,12,4);
    $tone_freq /= 10 if ($tone_type < 2);
    return $tone_freq;
}

sub set_gps {
    my $mode = shift;
    my $xmode = "00";
    $xmode = "00" if ($mode =~ /off/i);
    $xmode = "01" if ($mode =~ /on/i);
    &send2("fefe02801d4000${xmode}fd");
}

sub get_gps {
    &send("fefe02801d4000fd");
    my $msg = &recv("fefe80021d4000",0);
    my $mode = substr($msg,14,2);
    my $xmode = "Unknown $mode";
    $xmode = "OFF" if ($mode eq "00");
    $xmode = "ON" if ($mode eq "01");
    return $xmode;
}

sub set_gps_msg {
  my $msg = shift;
  my $msg = sprintf("%-20.20s",$msg);

  my $m_hex = unpack("H*",$msg);
  &send2("fefe02801d41${m_hex}fd");
}

sub get_gps_msg {
  &send("fefe02801d41fd");
  my $msg = &recv("fefe80021d41",0);
  my $m_text = pack("H*",substr($msg,12,40));
  return $m_text;
}

sub set_short_msg {
  my $msg = shift;
  my $msg = sprintf("%-32.32s",$msg);
  my $m_hex = unpack("H*",$msg);
  &send2("fefe02801ddf${m_hex}fd");
}

sub get_short_msg {
  &send("fefe02801ddffd");
  my $msg = &recv("fefe80021ddf",0);
  my $m_text = pack("H*",substr($msg,12,64));
  return $m_text;
} 

sub set_emr {
    my $mode = shift;
    my $xmode = "00";
    $xmode = "00" if ($mode =~ /off/i);
    $xmode = "01" if ($mode =~ /on/i);
    &send2("fefe02801dec${xmode}fd");
}

sub get_emr {
    &send("fefe02801decfd");
    my $msg = &recv("fefe80021dec",0);
    my $mode = substr($msg,12,2);
    my $xmode = "Unknown $mode";
    $xmode = "OFF" if ($mode eq "00");
    $xmode = "ON" if ($mode eq "01");
    return $xmode;
}

sub set_bk {
    my $mode = shift;
    my $xmode = "00";
    $xmode = "00" if ($mode =~ /off/i);
    $xmode = "01" if ($mode =~ /on/i);
    &send2("fefe02801d10${xmode}fd");
}

sub get_bk {
    &send("fefe02801d10fd");
    my $msg = &recv("fefe80021d10",0);
    my $mode = substr($msg,12,2);
    my $xmode = "Unknown $mode";
    $xmode = "OFF" if ($mode eq "00");
    $xmode = "ON" if ($mode eq "01");
    return $xmode;
}

sub set_backlight {
    my $mode = shift;
    my $xmode = "00";
    $xmode = "00" if ($mode =~ /off/i);
    $xmode = "01" if ($mode =~ /on/i);
    $xmode = "02" if ($mode =~ /auto/i);
    &send2("fefe01801a0510${xmode}fd");
}

sub get_backlight {
    &send("fefe01801a0510fd");
    my $msg = &recv("fefe80011a0510",0);
    my $mode = substr($msg,14,2);
    my $xmode = "Unknown $mode";
    $xmode = "OFF" if ($mode eq "00");
    $xmode = "ON" if ($mode eq "01");
    $xmode = "AUTO" if ($mode eq "02");
    return $xmode;
}

sub set_att {
    my $mode = shift;
    my $xmode = "00";
    $xmode = "00" if ($mode =~ /off/i);
    $xmode = "01" if ($mode =~ /on/i);
    &send2("fefe018011${xmode}fd");
}

sub get_att {
    &send("fefe018011fd");
    my $msg = &recv("fefe800111",0);
    my $mode = substr($msg,10,2);
    my $xmode = "Unknown $mode";
    $xmode = "OFF" if ($mode eq "00");
    $xmode = "ON" if ($mode eq "01");
    return $xmode;
}

sub get_battery {
    &send("fefe01801a040efd");
    my $msg = &recv("fefe80011a040e");
    my $bstat = &decode_battery($msg);
    return $bstat;
}

sub decode_battery {
    my $msg = shift;
    my $bcode = substr($msg,14,2);
    my $bstat = $bcode;
    $bstat = "External Power" if ($bstat eq "00");
    $bstat = "Battery Power" if ($bstat eq "01");
    $bstat = "Low Battery" if ($bstat eq "02");
    $bstat = "Battery Charging" if ($bstat eq "03");
    return $bstat;
}



sub set_sql {
    my $lvl = shift;
    my $band = &get_bandno;
    my $xlvl = "01";  #default to 'auto'
    # Note: level "00" does not open squelch, but sets it to auto in DV mode
    #       like "01" does.
    $xlvl = "00" if ($lvl =~ /open/i);
    $xlvl = "01" if ($lvl =~ /auto/i);
    if ($lvl =~ /^\d+$/) {
        $xlvl = sprintf("%02x",$lvl + 1);
    }
    &send2("fefe${band}801403${xlvl}fd");
}

sub get_sql {
    my $band = &get_bandno;
    &send("fefe${band}801403fd");
    my $msg = &recv("fefe80${band}1403");
    my $lvl = substr($msg,12,2);
    my $xlvl = "Unknown $lvl";
    if ($lvl eq "00") {
        $xlvl = "OPEN";
    }
    elsif ($lvl eq "01") {
        $xlvl = "AUTO";
    }
    else {
        $xlvl = hex($lvl) - 1;
    }
    return $xlvl;
}

sub set_vol {
    my $lvl = shift;
    # volume level 0-39
    my $xlvl = "10"; 
    if ($lvl >=0 && $lvl < 40) {
        $xlvl = sprintf("%02d",$lvl);
    }
    else {
        print "\nInvalid volume level: $lvl\n";
        return;
    }

    # Don't need to do both bands.
    #&send2("fefe0180140100${xlvl}fd"); #band 1
    &send2("fefe0280140100${xlvl}fd"); #band 2
}

sub get_vol {
    &send("fefe02801401fd");
    my $msg = &recv("fefe80021401");
    my $b_lvl = substr($msg,12,4);
    $b_lvl += 0;

#    &send("fefe01801401fd");
#    my $msg = &recv("fefe80011401");
#    my $a_lvl = substr($msg,12,4);
#    $a_lvl += 0;

#    return "Band A: $a_lvl   Band B: $b_lvl";
     return $b_lvl;
}

sub set_busy_lockout {
    my $mode = shift;
    my $xmode = "00";
    $xmode = "00" if ($mode =~ /off/i);
    $xmode = "01" if ($mode =~ /on/i);
    &send2("fefe01801a0549${xmode}fd");
}

sub get_busy_lockout {
    &send("fefe01801a0549fd");
    my $msg = &recv("fefe80011a0549",0);
    my $mode = substr($msg,14,2);
    my $xmode = "Unknown $mode";
    $xmode = "OFF" if ($mode eq "00");
    $xmode = "ON" if ($mode eq "01");
    return $xmode;
}

sub set_msg_prefix {
    my $pre = shift;
    if ( defined($pre) && length($pre) > 0) {
        $pre =~ s/"//g;
        $msg_prefix = $pre;
    }
}

sub set_band {
  my $band = shift;
  my $xband = "00";
  $xband = "00" if ($band =~ /^a/i);
  $xband = "01" if ($band =~ /^b/i);
  $xband = "00" if ($band =~ /^s/i);
  $xband = "01" if ($band =~ /^d/i);

  if ($band =~ /^a|b/i) {
    # select A/B band
    &send2("fefe01801a0401${xband}fd");
  }
  elsif ($band =~ /^s|d/i) {
    # select single/dual band mode
    &send2("fefe01801a0400${xband}fd");
  }
}

sub get_band {
  my $msg = &get_bandmsg;
  my $xband = &decode_band($msg);

  &send("fefe01801a0400fd");
  $msg = &recv("fefe80011a0400",0);
  my $xdual = &decode_band_mode($msg);

  return "$xdual mode on \"$xband\"";
}

# Get which band A/B is selected
sub get_bandmsg {
  &send("fefe01801a0401fd");
  my $msg = &recv("fefe80011a0401",0);
  return $msg;
}

# get the current band number (01/02) from the band request message reply
sub get_bandno {
  my $msg = &get_bandmsg;
  my $band = substr($msg,14,2);
  #convert to band number used in radio commands
  $band = sprintf("%02d",$band + 1);
  return $band;
}

sub decode_band {
  my $msg = shift;
  my $band = substr($msg,14,2);
  my $xband = $band;
  $xband = "A" if ($band eq "00");
  $xband = "B" if ($band eq "01");
  return $xband;
}

sub decode_band_mode {
  my $msg = shift;
  my $dual = substr($msg,14,2);
  my $xdual = $dual;
  $xdual = "Single Band" if ($dual eq "00");
  $xdual = "Dual Band" if ($dual eq "01");
  return $xdual;
}

# show help
sub do_help {
  my $subject = shift;
  my $version = '$RCSfile: drc,v $ $Revision: 1.22 $';
  my $htext = <<EOT;
$version
Usage: $0 [--port <port>] [--baud <baud>] [--debug] [--help]
       default port: /dev/ttyS0
       default baud: 38400

Operation Commands:
  /show att                     ; shows attenuator setting
  /show backlight               ; shows backlight setting
  /show band                    ; shows selected band and single/dual band mode
  /show battery                 ; shows Battery/Power status
  /show bk                      ; shows BK mode
  /show busylock                ; shows busy lockout setting (on/off)
  /show dsql                    ; shows digital/call-sign squelch setting
  /show dtcs                    ; shows current DTCS code
  /show dup                     ; shows current duplex setting
  /show emr                     ; shows EMR mode
  /show freq                    ; shows current operating frequency
  /show gps                     ; shows GPS mode
  /show mem                     ; shows current memory channel
  /show mode                    ; shows current operating mode (FM/AM/DV)
  /show msgpre                  ; shows message prefix
  /show mycall                  ; shows mycall memory and TX callsigns.
  /show offset                  ; shows current repeater offset value
  /show power                   ; shows output power level setting (0=Low 255=High)
  /show psave                   ; shows Power Saver mode (off, auto)
  /show rtone                   ; shows current repeater tone frequency
  /show rxcall                  ; shows RX callsign history
  /show sql                     ; shows current squelch level
  /show tone                    ; shows current tone/tone squelch setting
  /show tsql                    ; shows current TSQL (CTCSS) frequency
  /show txcall                  ; shows TX callsign settings.
  /show vfo                     ; shows current operating frequency
  /show vol                     ; shows volume level
  /set att [on|off]             ; set attenuator on/off
  /set backlight [on|off|auto]  ; set backlight mode
  /set band [a|b|single|dual]   ; set band or single/dual band mode
  /set bk [on|off]              ; set Break-In communication mode
  /set busylock [on|off]        ; set Busy Lockout on or off.
  /set dsql [off|DSQL [code]|CSQL]
  /set dtcs <nn>                ; set DTCS code (00-99)
  /set dup [+|-|OFF]            ; set duplex mode
  /set emr [on|off]             ; set EMR communication mode
  /set gps [on|off]             ; set GPS mode
  /set mem   <[A-Z]nn | nnn>    ; set memory channel (i.e. B23 -or- 200)
  /set msgpre <text>            ; Prefix each outgoing message with <text>
  /set mode [FM|FM-N|AM|DV      ; set modulation mode
  /set mycall <n>               ; selects mycall memory <n>
  /set mycall <n> -             ; clears mycall memory <n>
  /set mycall <n> <call>[/xxxx] ; sets mycall memory <n>
  /set offset <n>               ; set repeater offset in Mhz
  /set power <n>|L|H            ; set output power 0=low 255=high, L=low H=high
  /set psave [off|auto]         ; set Power Saver mode
  /set rtone <n>                ; set Repeater Tone frequency
  /set sql <n>|auto|open        ; set squelch level n=0-9, "auto" or "open"
  /set tone [TONE|TSQL|DTCS]    ; set Tone/Tone squelch mode
  /set tsql <n>                 ; set TSQL (CTCSS) frequency
  /set txcall [[<rpt2>].<rpt1>.][<urcall>]
  /set vfo [<freq>|mem]         ; set radio in VFO mode at the specified frequency.
                                ; 'mem' copies current memory to VFO.
  /set volume <n>               ; set audio volume level (0-39)
  /heard                        ; same as /show rxcall
  /say <text>                   ; send <text>
  /exit
  /quit
EOT

  my $line_no = 0;
  my $max_line = `echo \$LINES`;
  chomp $max_line;
  if ($max_line > 0 ) {
    $max_line--;
  }
  else {
    $max_line = 25;
  }
  my @help_text;
  if( $subject ) {
    @help_text = grep(/$subject/i,split(/\n/,$htext));
  }
  else {
    @help_text = split(/\n/,$htext);
  }
  print "\n";
  for my $line (@help_text) {
    print "$line\n";
    $line_no++;
    if ($line_no == $max_line) {
      print "--MORE--";
      my $imp = "";
      my $sleep_cnt = 0;
      while (sysread(CMD,$imp,1) == 0) {
        &ms_sleep(0.1);
        $sleep_cnt++;
        if($sleep_cnt % 10 == 0) {
          # Keep radio comm link alive.
          # Don't let the comm link be idle for more than 3 seconds or it resets.
          # The threshold seems to be a bit longer than 3 seconds, but we
          # use 3 to be safe.
          # Note: I've changed this to 1 second to increase response time to
          #       unsolicited messages from the radio.
          #
          &rx_poll if($radio_open);
        }
      }
      &rx_poll if($radio_open);
      print "\r        \r";
      last if ($imp eq "q");
      if ($imp eq "j") {
        $line_no = $max_line - 1;  # show one more line only
      }
      else {
        $line_no = 0;
      }
    }
  }
  print STDERR "\nExit Help\n" if ($debug);
}



#
# $Log: drc,v $
# Revision 1.22  2009-05-26 20:08:29  bruce
# Update comments, add another debugging/investiagion command /rcmdx
# which will translate a response to ASCII.
#
# Revision 1.21  2009-04-03 20:42:26  bruce
# Increase number of 'fe's in interrogate sequence as the radio wasn't always responding
# the first time.
# Decrease polling time.
# Eliminate sleeps that were occuring between streems of data coming from the radio.
# Add gps and gpsmsg settings.
# Debugging and be dynamically turned on and off.
# /rcmd can send raw commands to the radio for debugging/exploring the protocol
# and capabilities.
#
# Revision 1.20  2009-03-20 22:20:56  bruce
# Document GPS messages
# If an expected ACK/NAK was never received, the program would become unresponsive
# Add command to copy current memory to VFO.
#
# Revision 1.19  2009-03-17 22:06:36  bruce
# Updated TODO list.
# Added Power status handling.
# Can now recover if communication lost with radio.
#
# Revision 1.18  2009-03-10 20:51:24  bruce
# Save/restore Power saver mode.
# Force Power Saver mode to off while program is running.
#
# Revision 1.17  2009-03-06 21:04:21  bruce
# Enhance help so that a help subject can be specified.
#
# Revision 1.16  2009-03-06 18:21:01  bruce
# Fix commands to be dual band aware.  On commands that are band sensitive,
# make sure they operate on the currently selected band. Previously, many
# commands were set to operate only on the "B" band (where DV is available,
# and the only band where text messaging works).
#
# Revision 1.15  2009-03-05 23:14:02  bruce
# Setting memory channel now works for A and B bands rather than just B.
#
# Revision 1.14  2009-03-05 23:07:04  bruce
# Fix get_s_meter routine, was incorrectly named get_sql.
#
# Revision 1.13  2009-03-05 22:17:25  bruce
# Fix help when invoked from command line (and radio not open).
#
# Revision 1.12  2009-03-05 22:05:34  bruce
# Add command to set band A/B dual/single.
# Show more detail about selected memory channel
# Fix radio polling during help display.
#
# Revision 1.11  2009-03-04 16:32:53  bruce
# Fix Revsion output in help
#
# Revision 1.10  2009-03-04 16:28:42  bruce
# Add CVS Revision to Help output.
#
# Revision 1.9  2009-03-04 16:22:19  bruce
# Modified help output.
# Added command line options to set port, baud
#
# Revision 1.8  2008-01-18 22:59:05  bruce
# Add ability to set an outgoing message prefix.
# Allow 'h' and 'l' for High and Low power setting values.
# Change fromat of timestamp and ID messages.
#
# Revision 1.7  2008-01-18 06:06:13  bruce
# Only set PTT to AUTO at start of program, and set it back upon exit.
# Leaving PTT on AUTO seems to fix the incomplete messages being sent.
# Queue incoming message until we get an end-of-line. Then display it.
#
# Revision 1.6  2008-01-16 06:25:41  bruce
# Add settings of various tone and squelch modes.
# Add setting of GPS and backlight.
# Add setting of EMR and BK modes.
#
# Revision 1.5  2008-01-15 22:50:22  bruce
# Many more settings can now be made.
# Can switch between VFO/Memory/MemoryBank modes
#
# Revision 1.4  2008-01-04 22:35:54  bruce
# Can now set mycall dynamically.
# Fix so ^C will kill program.
# Add /help command
#
# Revision 1.3  2008-01-04 20:42:52  bruce
# More enhancements and bug fixes.
# Add /show command.
# Fix get_mycall, was off by one.
# Fix sending of long messages by ignoring any data before a data preable
#     when receiving data from radio (looking for ACK, or a status reply).
#
# Revision 1.2  2008-01-04 06:39:21  bruce
# Various fixes and enhancements.
# Add /mycalls command
# Long messages need to be sent in 32 byte segments.
#
# Revision 1.1  2008-01-04 04:35:09  bruce
# Initial checkin.
#
#
